#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "BituminousRoofingRangeV1",
    "SinglePlyEPDMRangeV1",
    "SinglePlyKEERangeV1",
    "SinglePlyOtherRangeV1",
    "SinglePlyPolyurethaneRangeV1",
    "SinglePlyPVCRangeV1",
    "SinglePlyTPORangeV1",
    "BlanketInsulationRangeV1",
    "BlownInsulationRangeV1",
    "BoardInsulationRangeV1",
    "FoamedInPlaceRangeV1",
    "SprayedInsulationRangeV1",
    "AirBarriersRangeV1",
    "MembraneRoofingRangeV1",
    "InsulationRangeV1",
    "DampproofingAndWaterproofingRangeV1",
    "FlashingAndSheetMetalRangeV1",
    "JointProtectionRangeV1",
    "RoofCoverBoardsRangeV1",
    "SteepSlopeRoofingRangeV1",
    "WeatherBarriersRangeV1",
    "ThermalMoistureProtectionRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.common import RangeFloat, RangeRatioFloat
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import (
    FoamType,
    InsulatingMaterial,
    InsulationIntendedApplication,
    MembraneRoofingReinforcement,
    RoofCoverBoardsFacing,
    RoofCoverBoardsMaterial,
)
from openepd.model.validation.quantity import AmountRangeLengthMm, AmountRangePressureMpa


class BituminousRoofingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Bituminous roofing.

    Premanufactured membrane roofing sheets consisting of asphalt, reinforcing layers, and in some cases a surfacing.

    Range version.
    """

    _EXT_VERSION = "1.0"


class SinglePlyEPDMRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Ethylene propylene diene monomer (EPDM) rubber membrane.

    Range version.
    """

    _EXT_VERSION = "1.0"


class SinglePlyKEERangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Ketone Ethylene Ester (KEE) roof membranes.

    Consist of PVC resin and KEE plasticizer, with KEE exceeding 50% of the polymer content by weight.

    Range version.
    """

    _EXT_VERSION = "1.0"


class SinglePlyOtherRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Single ply other performance specification.

    Range version.
    """

    _EXT_VERSION = "1.0"


class SinglePlyPolyurethaneRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Polyurethane liquid for flat roof waterproofing.

    Range version.
    """

    _EXT_VERSION = "1.0"


class SinglePlyPVCRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Polyvinyl chloride (PVC) thermoplastic membrane.

    Range version.
    """

    _EXT_VERSION = "1.0"


class SinglePlyTPORangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Thermoplastic Polyolefin (TPO) membrane.

    Range version.
    """

    _EXT_VERSION = "1.0"


class BlanketInsulationRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Non-rigid insulation batts, blankets, and rolls.

    Range version.
    """

    _EXT_VERSION = "1.0"


class BlownInsulationRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Loose-fill insulation for blow-in or closed cavity applications.

    Range version.
    """

    _EXT_VERSION = "1.0"


class BoardInsulationRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Rigid insulation products including rigid foams, wood fiberboard insulation, and rigid mineral wool boards.

    Range version.
    """

    _EXT_VERSION = "1.0"

    compressive_strength: AmountRangePressureMpa | None = pyd.Field(default=None, description="")


class FoamedInPlaceRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Open and closed cell spray foam insulation.

    Range version.
    """

    _EXT_VERSION = "1.0"

    foam_type: list[FoamType] | None = pyd.Field(default=None, description="")


class SprayedInsulationRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Spray-on insulation, such as spray-on cellulose.

    Foaming sprays are categorized separately under foamed-in-place.

    Range version.
    """

    _EXT_VERSION = "1.0"


class AirBarriersRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Air Infiltration Barrier.

    Range version.
    """

    _EXT_VERSION = "1.0"


class MembraneRoofingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Membrane roofing.

    Built-up bituminous, modified bituminous, elastomeric, thermoplastic, fluid-applied, and hot-applied rubberized
    asphalt membrane roofing.

    Range version.
    """

    _EXT_VERSION = "1.0"

    thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    sri: RangeFloat | None = pyd.Field(default=None, description="")
    total_recycled_content: RangeRatioFloat | None = pyd.Field(default=None, description="")
    post_consumer_recycled_content: RangeRatioFloat | None = pyd.Field(default=None, description="")
    reinforcement: list[MembraneRoofingReinforcement] | None = pyd.Field(default=None, description="")
    felt_backing: bool | None = pyd.Field(default=None, description="")
    nsf347: bool | None = pyd.Field(default=None, description="")
    vantage_vinyl: bool | None = pyd.Field(default=None, description="")
    BituminousRoofing: BituminousRoofingRangeV1 | None = None
    SinglePlyEPDM: SinglePlyEPDMRangeV1 | None = None
    SinglePlyKEE: SinglePlyKEERangeV1 | None = None
    SinglePlyOther: SinglePlyOtherRangeV1 | None = None
    SinglePlyPolyurethane: SinglePlyPolyurethaneRangeV1 | None = None
    SinglePlyPVC: SinglePlyPVCRangeV1 | None = None
    SinglePlyTPO: SinglePlyTPORangeV1 | None = None


class InsulationRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Thermal insulation materials for use in construction.

    Range version.
    """

    _EXT_VERSION = "1.0"

    r_value: RangeFloat | None = pyd.Field(default=None, description="")
    material: list[InsulatingMaterial] | None = pyd.Field(default=None, description="")
    intended_application: list[InsulationIntendedApplication] | None = pyd.Field(default=None, description="")
    thickness_per_declared_unit: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    BlanketInsulation: BlanketInsulationRangeV1 | None = None
    BlownInsulation: BlownInsulationRangeV1 | None = None
    BoardInsulation: BoardInsulationRangeV1 | None = None
    FoamedInPlace: FoamedInPlaceRangeV1 | None = None
    SprayedInsulation: SprayedInsulationRangeV1 | None = None


class DampproofingAndWaterproofingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Dampproofing and waterproofing.

    Dampproofing, and built-up bituminous, sheet, fluid-applied, cementitious, reactive, and bentonite waterproofing.

    Range version.
    """

    _EXT_VERSION = "1.0"


class FlashingAndSheetMetalRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Exposed sheet metal items, typically for drainage.

    Range version.
    """

    _EXT_VERSION = "1.0"


class JointProtectionRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Preformed joint seals and sealants, expansion control joint cover assemblies.

    Range version.
    """

    _EXT_VERSION = "1.0"


class RoofCoverBoardsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Boards installed between the insulation and membrane layers on a roof system.

    It provides additional durability, fire protection, thermal, and vapor performance to a roof system, especially
    in low-slope, high foot traffic applications.

    Range version.
    """

    _EXT_VERSION = "1.0"

    material: list[RoofCoverBoardsMaterial] | None = pyd.Field(default=None, description="")
    facing: list[RoofCoverBoardsFacing] | None = pyd.Field(default=None, description="")
    thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")


class SteepSlopeRoofingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Roofing materials typically for slopes of 3:12 and greater.

    Range version.
    """

    _EXT_VERSION = "1.0"


class WeatherBarriersRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Vapor retarders and sheet or membrane air barriers.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ThermalMoistureProtectionRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Thermal moisture protection.

    Broad category of materials whose function is to provide moisture and thermal protection between spaces (e.g.,
    between the exterior and interior of a building).

    Range version.
    """

    _EXT_VERSION = "1.0"

    AirBarriers: AirBarriersRangeV1 | None = None
    MembraneRoofing: MembraneRoofingRangeV1 | None = None
    Insulation: InsulationRangeV1 | None = None
    DampproofingAndWaterproofing: DampproofingAndWaterproofingRangeV1 | None = None
    FlashingAndSheetMetal: FlashingAndSheetMetalRangeV1 | None = None
    JointProtection: JointProtectionRangeV1 | None = None
    RoofCoverBoards: RoofCoverBoardsRangeV1 | None = None
    SteepSlopeRoofing: SteepSlopeRoofingRangeV1 | None = None
    WeatherBarriers: WeatherBarriersRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "AluminiumBilletsRangeV1",
    "AluminiumExtrusionsRangeV1",
    "AluminiumSheetGoodsRangeV1",
    "AluminiumSuspensionAssemblyRangeV1",
    "AluminiumRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import AluminiumAlloy


class AluminiumBilletsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Cast Aluminium ingots or billets for use in manufacturing more finished products.

    Range version.
    """

    _EXT_VERSION = "1.0"


class AluminiumExtrusionsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Extruded aluminum products used in construction.

    Includes range of finish options including mill finish, painted, and anodized.


    Range version.
    """

    _EXT_VERSION = "1.0"

    thermally_improved: bool | None = pyd.Field(default=None, description="")


class AluminiumSheetGoodsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Rolled and/or Stamped Aluminium coil or sheets, often used in flashing, trim, panels, and deck.

    Range version.
    """

    _EXT_VERSION = "1.0"


class AluminiumSuspensionAssemblyRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Aluminum suspension assemblies for acoustical ceiling systems.

    Range version.
    """

    _EXT_VERSION = "1.0"


class AluminiumRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Broad category for construction materials made primarily from Aluminium and its alloys.

    Range version.
    """

    _EXT_VERSION = "1.0"

    alloy: list[AluminiumAlloy] | None = pyd.Field(default=None, description="")
    anodized: bool | None = pyd.Field(default=None, description="")
    painted: bool | None = pyd.Field(default=None, description="")
    AluminiumBillets: AluminiumBilletsRangeV1 | None = None
    AluminiumExtrusions: AluminiumExtrusionsRangeV1 | None = None
    AluminiumSheetGoods: AluminiumSheetGoodsRangeV1 | None = None
    AluminiumSuspensionAssembly: AluminiumSuspensionAssemblyRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import ClassVar

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema
from openepd.model.specs.enums import SteelComposition
from openepd.model.specs.singular import BaseCompatibilitySpec
from openepd.model.specs.singular.steel import SteelMakingRoute
from openepd.model.standard import Standard
from openepd.model.validation.quantity import PressureMPaStr


class SteelOldOptions(BaseOpenEpdSchema):
    """Legacy Steel options model."""

    cold_finished: bool | None = pyd.Field(default=None, description="Cold Finished", example=True)
    galvanized: bool | None = pyd.Field(default=None, description="Galvanized", example=True)
    epoxy: bool | None = pyd.Field(default=None, description="Epoxy Coated", example=True)
    steel_fabricated: bool | None = pyd.Field(default=None, example=True, description="Fabricated")


class SteelOldSpec(BaseCompatibilitySpec):
    """Legacy Steel spec."""

    COMPATIBILITY_SPECS_KEY_OLD: ClassVar[str] = "steel"
    COMPATIBILITY_SPECS_KEY_NEW: ClassVar[str] = "Steel"
    COMPATIBILITY_MAPPING: ClassVar[dict[str, str]] = {
        "steel.steel_composition": "Steel.composition",
        "steel.Fy": "Steel.yield_tensile_str",
        "steel.making_route": "Steel.making_route",
        "steel.ASTM": "Steel.astm_standards",
        "steel.EN": "Steel.en_standards",
        "steel.SAE": "Steel.sae_standards",
        "steel.options.cold_finished": "Steel.cold_finished",
        "steel.options.galvanized": "Steel.galvanized",
        "steel.options.epoxy": "Steel.RebarSteel.epoxy_coated",  # moved to sub-spec
        "steel.options.steel_fabricated": "Steel.RebarSteel.fabricated",
        # last one is tricky as fabricated might be in multiple cases; thus limited support
    }
    form_factor: str | None = pyd.Field(default=None, description="Product's form factor, read-only.")
    steel_composition: SteelComposition | None = pyd.Field(
        default=None,
        description="Basic chemical composition. Generally the ASTM or EN grade is a subcategory of one of these.",
        example="Carbon",
    )
    Fy: PressureMPaStr | None = pyd.Field(
        default=None,
        description="Minimum Yield Strength",
        example="2000 psi",
    )
    making_route: SteelMakingRoute | None = pyd.Field(
        default=None, description="List of true/false properties for steelmaking route"
    )
    ASTM: list[Standard] | None = pyd.Field(
        default=None, description="ASTM standard(s) to which this product complies."
    )
    SAE: list[Standard] | None = pyd.Field(
        default=None, description="AISA/SAE standard(s) to which this product complies."
    )
    EN: list[Standard] | None = pyd.Field(default=None, description="EN 10027 number(s).")
    options: SteelOldOptions | None = pyd.Field(default=None, description="List of true/false properties")



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "AccessFlooringRangeV1",
    "CarpetRangeV1",
    "LaminateRangeV1",
    "OtherFlooringRangeV1",
    "ResilientFlooringRangeV1",
    "WallBaseRangeV1",
    "WoodFlooringRangeV1",
    "AcousticalCeilingsRangeV1",
    "CeramicTileRangeV1",
    "GaugedTileRangeV1",
    "GlassTileRangeV1",
    "GypsumSupportsRangeV1",
    "FlooringRangeV1",
    "CeilingPanelRangeV1",
    "BackingAndUnderlayRangeV1",
    "CementBoardRangeV1",
    "TilingRangeV1",
    "DeckingBoardsRangeV1",
    "GlassFiberReinforcedGypsumFabricationsRangeV1",
    "GypsumRangeV1",
    "MirrorsRangeV1",
    "PaintByMassRangeV1",
    "PaintByVolumeRangeV1",
    "PaintByAreaRangeV1",
    "PaintingAndCoatingRangeV1",
    "WallFinishesRangeV1",
    "PlasterRangeV1",
    "FinishesRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.common import RangeAmount, RangeFloat, RangeInt, RangeRatioFloat
from openepd.model.org import OrgRef
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import (
    AccessFlooringCoreMaterial,
    AccessFlooringFinishMaterial,
    AccessFlooringSeismicRating,
    AccessFlooringStringers,
    AllFabrication,
    CarpetFormFactor,
    CarpetIntendedApplication,
    CarpetManufactureType,
    CarpetYarnType,
    CeilingPanelCoreMaterial,
    CeilingPanelFireRating,
    CementBoardThickness,
    DeckingBoardMaterial,
    GypsumFacing,
    GypsumFireRating,
    GypsumThickness,
    PlasterComposition,
    ResilientFlooringFormFactor,
    ResilientFlooringMaterial,
    ResilientFlooringThickness,
    SawnTimberSpecies,
    VinylSheetConstruction,
    WallBaseMaterial,
    WoodFlooringFabrication,
    WoodFlooringTimberSpecies,
)
from openepd.model.validation.quantity import (
    AmountRangeForce,
    AmountRangeGWP,
    AmountRangeLengthMm,
    AmountRangePressureMpa,
    AmountRangeRFactor,
    AmountRangeYarnWeight,
)


class AccessFlooringRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Elevated floor system built on top of concrete slab surface.

    It thereby creates a hidden void between the two floors that is used for the passage of mechanical and electrical
    services. The system consists of panels, stringers, and pedestals.

    Range version.
    """

    _EXT_VERSION = "1.0"

    core_material: list[AccessFlooringCoreMaterial] | None = pyd.Field(default=None, description="")
    finish_material: list[AccessFlooringFinishMaterial] | None = pyd.Field(default=None, description="")
    stringers: list[AccessFlooringStringers] | None = pyd.Field(default=None, description="")
    seismic_rating: list[AccessFlooringSeismicRating] | None = pyd.Field(default=None, description="")
    magnetically_attached_finish: bool | None = pyd.Field(default=None, description="")
    permanent_finish: bool | None = pyd.Field(default=None, description="")
    drylay: bool | None = pyd.Field(default=None, description="")
    adjustable_height: bool | None = pyd.Field(default=None, description="")
    fixed_height: bool | None = pyd.Field(default=None, description="")
    finished_floor_height: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    panel_thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    concentrated_load: AmountRangePressureMpa | None = pyd.Field(default=None, description="")
    uniform_load: AmountRangePressureMpa | None = pyd.Field(default=None, description="")
    rolling_load_10_pass: AmountRangeForce | None = pyd.Field(default=None, description="")
    rolling_load_10000_pass: AmountRangeForce | None = pyd.Field(default=None, description="")


class CarpetRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Textile Floor Coverings.

    Range version.
    """

    _EXT_VERSION = "1.0"

    length: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    width: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    intended_application: list[CarpetIntendedApplication] | None = pyd.Field(default=None, description="")
    manufacture_type: list[CarpetManufactureType] | None = pyd.Field(default=None, description="")
    form_factor: list[CarpetFormFactor] | None = pyd.Field(default=None, description="")
    yarn_weight: AmountRangeYarnWeight | None = pyd.Field(default=None, description="")
    yarn_type: list[CarpetYarnType] | None = pyd.Field(default=None, description="")
    fire_radiant_panel_rating_astme648: str | None = pyd.Field(default=None, description="")
    fire_smoke_density_rating_astme648: str | None = pyd.Field(default=None, description="")
    voc_emissions: str | None = pyd.Field(default=None, description="")
    cushioned: bool | None = pyd.Field(default=None, description="")
    bleachable: bool | None = pyd.Field(default=None, description="")
    gwp_factor_base: AmountRangeGWP | None = pyd.Field(default=None, description="")
    gwp_factor_yarn: AmountRangeGWP | None = pyd.Field(default=None, description="")


class LaminateRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Laminate flooring.

    Range version.
    """

    _EXT_VERSION = "1.0"


class OtherFlooringRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Other not yet classified kinds of flooring.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ResilientFlooringRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Resilient floor products.

    Includes vinyl, rubber, linoleum, composition cork, etc. in modular square or rectangle shapes.

    Range version.
    """

    _EXT_VERSION = "1.0"

    length: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    width: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    form_factor: list[ResilientFlooringFormFactor] | None = pyd.Field(default=None, description="")
    material: list[ResilientFlooringMaterial] | None = pyd.Field(default=None, description="")
    sheet_construction: list[VinylSheetConstruction] | None = pyd.Field(default=None, description="")
    wear_layer: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    delta_iic: RangeFloat | None = pyd.Field(default=None, description="")
    thickness: list[ResilientFlooringThickness] | None = pyd.Field(default=None, description="")
    sport_flooring: bool | None = pyd.Field(default=None, description="")
    conductive_flooring: bool | None = pyd.Field(default=None, description="")
    zwtl: bool | None = pyd.Field(default=None, description="")
    floor_score: bool | None = pyd.Field(default=None, description="")
    nsf332: bool | None = pyd.Field(default=None, description="")


class WallBaseRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Wall base made to help cover gaps between wall and vinyl, rubber, wood, or tile flooring.

    Range version.
    """

    _EXT_VERSION = "1.0"

    wall_base_material: list[WallBaseMaterial] | None = pyd.Field(default=None, description="")


class WoodFlooringRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Wood flooring for interior applications.

    Includes hardwood strip and plank flooring, engineered hardwood flooring, wood parquet flooring, coordinated
    transitions, and molding pieces.

    Range version.
    """

    _EXT_VERSION = "1.0"

    forest_practices_certifiers: list[OrgRef] | None = None
    thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    timber_species: list[WoodFlooringTimberSpecies] | None = pyd.Field(default=None, description="")
    fabrication: list[WoodFlooringFabrication] | None = pyd.Field(default=None, description="")


class AcousticalCeilingsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Acoustical ceiling panels.

    Range version.
    """

    _EXT_VERSION = "1.0"

    thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")


class CeramicTileRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Ceramic tiles, including porcelain, quarry, pressed floor tile, wall tile, mosaic tile, etc.

    Range version.
    """

    _EXT_VERSION = "1.0"

    porcelain: bool | None = pyd.Field(
        default=None, description="A dense and durable ceramic tile made from fine porcelain clay."
    )
    quarry: bool | None = pyd.Field(
        default=None,
        description="A type of unglazed ceramic tile made from natural clay with a slightly rough texture.",
    )
    pressed_floor_tile: bool | None = pyd.Field(
        default=None,
        description="A durable and low-maintenance type of tile made by compressing clay or other materials at high pressure.",
    )
    wall_tile: bool | None = pyd.Field(
        default=None,
        description="A decorative tile designed for use on vertical surfaces such as walls or backsplashes.",
    )
    mosaic_tile: bool | None = pyd.Field(
        default=None,
        description="A small decorative tile made of glass, stone, or ceramic, arranged in a pattern to create a design.",
    )
    specialty: bool | None = pyd.Field(
        default=None,
        description="A unique and customized type of tile, often made from unconventional materials or with specialized designs or finishes.",
    )


class GaugedTileRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Specially manufactured porcelain tile with extra-large panels/slabs.

    Manufactured to a specific thickness ranging from 2-20 mm.

    Range version.
    """

    _EXT_VERSION = "1.0"

    tile_panels: bool | None = pyd.Field(
        default=None,
        description="Large-format porcelain or natural stone tiles that are typically over 3 feet in length and width, designed for use in floor and wall installations to create a seamless and uninterrupted appearance.",
    )
    tile_pavers: bool | None = pyd.Field(
        default=None,
        description="Thick and durable porcelain or natural stone tiles that are commonly used in outdoor applications, such as patios, walkways, and driveways, due to their high resistance to weather and wear.",
    )


class GlassTileRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Glass Tiles.

    Range version.
    """

    _EXT_VERSION = "1.0"

    regular: bool | None = pyd.Field(
        default=None,
        description="Glass tile that is typically square or rectangular in shape, and used for a variety of decorative applications, such as kitchen backsplashes, shower walls, and accent borders.",
    )
    glass_mosaic: bool | None = pyd.Field(
        default=None,
        description="A small, decorative glass tile made in a variety of shapes and colors, used for intricate designs and patterns on walls, floors, and other surfaces.",
    )
    miniature_mosaic: bool | None = pyd.Field(
        default=None,
        description="Glass mosaic tile that is smaller in size than regular glass mosaic tile, often used for intricate details and designs in backsplashes, shower walls, and decorative accents.",
    )
    large_format: bool | None = pyd.Field(
        default=None,
        description="Glass tile that is larger in size than regular glass tile, often used to create a dramatic and modern effect in commercial and residential spaces.",
    )


class GypsumSupportsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Supports for suspended and furred gypsum wall and ceiling assemblies.

    Range version.
    """

    _EXT_VERSION = "1.0"


class FlooringRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    General category - finishes for floors.

    Range version.
    """

    _EXT_VERSION = "1.0"

    AccessFlooring: AccessFlooringRangeV1 | None = None
    Carpet: CarpetRangeV1 | None = None
    Laminate: LaminateRangeV1 | None = None
    OtherFlooring: OtherFlooringRangeV1 | None = None
    ResilientFlooring: ResilientFlooringRangeV1 | None = None
    WallBase: WallBaseRangeV1 | None = None
    WoodFlooring: WoodFlooringRangeV1 | None = None


class CeilingPanelRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Acoustical and other specialty ceiling panels.

    Range version.
    """

    _EXT_VERSION = "1.0"

    fire_rating: list[CeilingPanelFireRating] | None = pyd.Field(default=None, description="")
    core_material: list[CeilingPanelCoreMaterial] | None = pyd.Field(default=None, description="")
    nrc: RangeRatioFloat | None = pyd.Field(
        default=None, description="Noise Reduction Coefficient (NRC) or Sound Absorbtion Average (SAA) per ASTM C423"
    )
    cac: RangeInt | None = pyd.Field(default=None, description="Ceiling Attenuation Class (CAC) per ASTM E1414")
    AcousticalCeilings: AcousticalCeilingsRangeV1 | None = None


class BackingAndUnderlayRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Cementitious, glass-mat faced gypsum, and fibered gypsum backing boards to support finish materials.

    Range version.
    """

    _EXT_VERSION = "1.0"


class CementBoardRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Hard cementitious boards, typically used as a tile backer.

    Range version.
    """

    _EXT_VERSION = "1.0"

    thickness: list[CementBoardThickness] | None = pyd.Field(default=None, description="")


class TilingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Decorative building materials that includes a variety of ceramic, porcelain, and glass tiles.

    Used for covering and enhancing surfaces such as floors, walls, and countertops.

    Range version.
    """

    _EXT_VERSION = "1.0"

    thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    flooring: bool | None = pyd.Field(default=None, description="Tiling intended for walking.")
    wall_finish: bool | None = pyd.Field(
        default=None,
        description="A decorative tile designed for use on vertical surfaces such as walls or backsplashes.",
    )
    cladding: bool | None = pyd.Field(
        default=None,
        description="Tiling for exterior use, primarily used for the walls of buildings and structures, providing a protective and decorative layer that enhances the aesthetic appearance and weather resistance of the underlying structure.",
    )
    other: bool | None = pyd.Field(
        default=None, description="Tiling used as countertops, ceilings, furnishings, hardscapes etc."
    )
    residential_only: bool | None = pyd.Field(
        default=None,
        description="All commercial tile can also be used in residential applications, but the opposite may not be true. This selection allows to filter out tiling that is not intended for commercial applications.",
    )
    reinforced: bool | None = pyd.Field(
        default=None, description="Steel-reinforced ceramic tiles or tiles with other special reinforcing technology."
    )
    total_recycled_content: RangeRatioFloat | None = pyd.Field(
        default=None,
        description="Proportion of this product that is sourced from recycled content. Pre-consumer recycling is given a 50% weighting, 100% for post-consumer, by mass.",
    )
    post_consumer_recycled_content: RangeRatioFloat | None = pyd.Field(
        default=None,
        description="Proportion of this product that is sourced from post-consumer recycled content, by mass.",
    )
    CeramicTile: CeramicTileRangeV1 | None = None
    GaugedTile: GaugedTileRangeV1 | None = None
    GlassTile: GlassTileRangeV1 | None = None


class DeckingBoardsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Decking boards provide the finished surface of a deck and support the weight of people and furniture.

    Range version.
    """

    _EXT_VERSION = "1.0"

    forest_practices_certifiers: list[OrgRef] | None = None
    timber_species: list[SawnTimberSpecies] | None = pyd.Field(default=None, description="")
    fabrication: list[AllFabrication] | None = pyd.Field(default=None, description="")
    weather_exposed: bool | None = pyd.Field(default=None, description="")
    fire_retardant: bool | None = pyd.Field(default=None, description="")
    decay_resistant: bool | None = pyd.Field(default=None, description="")
    material: list[DeckingBoardMaterial] | None = pyd.Field(default=None, description="")


class GlassFiberReinforcedGypsumFabricationsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Gypsum with integrated glass fiber reinforcement, which may be fabricated in complex shapes or as a board.

    Range version.
    """

    _EXT_VERSION = "1.0"


class GypsumRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Gypsum board used for interior walls, ceilings, and similar applications.

    Range version.
    """

    _EXT_VERSION = "1.1"

    fire_rating: list[GypsumFireRating] | None = pyd.Field(default=None, description="")
    thickness: list[GypsumThickness] | None = pyd.Field(default=None, description="")
    facing: list[GypsumFacing] | None = pyd.Field(default=None, description="")
    r_factor: AmountRangeRFactor | None = pyd.Field(default=None, description="")
    flame_spread_astm_e84: RangeInt | None = pyd.Field(default=None, description="")
    smoke_production_astm_e84: RangeInt | None = pyd.Field(default=None, description="")
    surface_abrasion_d4977: RangeInt | None = pyd.Field(default=None, description="")
    indentation_d5420: RangeInt | None = pyd.Field(default=None, description="")
    soft_body_impact_e695: RangeInt | None = pyd.Field(default=None, description="")
    hard_body_impact_c1929: RangeInt | None = pyd.Field(default=None, description="")
    mold_resistant: bool | None = pyd.Field(default=None, description="")
    foil_backing: bool | None = pyd.Field(default=None, description="")
    moisture_resistant: bool | None = pyd.Field(default=None, description="")
    abuse_resistant: bool | None = pyd.Field(default=None, description="")
    GypsumSupports: GypsumSupportsRangeV1 | None = None


class MirrorsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Mirrors.

    Range version.
    """

    _EXT_VERSION = "1.0"


class PaintByMassRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Paintings and coatings by mass.

    Expected declared unit for products is mass of ready to use product.

    Range version.
    """

    _EXT_VERSION = "1.0"


class PaintByVolumeRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Paintings and coatings by volume.

    Expecting declared unit for products is volume of ready to use product.

    Range version.
    """

    _EXT_VERSION = "1.0"


class PaintByAreaRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Paintings and coatings by area.

    Expected declared unit is area of host surface covered by the product.

    Range version.
    """

    _EXT_VERSION = "1.0"


class PaintingAndCoatingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Paintings and coatings.

    Range version.
    """

    _EXT_VERSION = "1.1"

    PaintByMass: PaintByMassRangeV1 | None = None
    PaintByVolume: PaintByVolumeRangeV1 | None = None
    PaintByArea: PaintByAreaRangeV1 | None = None


class WallFinishesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Interior wall coverings including fabric, textile, wood, stone, and metal products.

    Range version.
    """

    _EXT_VERSION = "1.0"

    thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")


class PlasterRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Plaster, Stucco, & Render.

    Used for the protective or decorative coating of walls and ceilings and for
    moulding and casting decorative elements. These are typically gypsum-, lime-,
    or cement-based. Products in this category refer to dry mix.

    Range version.
    """

    _EXT_VERSION = "1.0"

    composition: list[PlasterComposition] | None = pyd.Field(default=None, description="")
    application_rate: RangeAmount | None = pyd.Field(
        default=None, description="Typical or reference amount of material covering a unit of a host surface."
    )


class FinishesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    General category - finishes for interior ceilings, floors, walls.

    Range version.
    """

    _EXT_VERSION = "1.1"

    Flooring: FlooringRangeV1 | None = None
    CeilingPanel: CeilingPanelRangeV1 | None = None
    BackingAndUnderlay: BackingAndUnderlayRangeV1 | None = None
    CementBoard: CementBoardRangeV1 | None = None
    Tiling: TilingRangeV1 | None = None
    DeckingBoards: DeckingBoardsRangeV1 | None = None
    GlassFiberReinforcedGypsumFabrications: GlassFiberReinforcedGypsumFabricationsRangeV1 | None = None
    Gypsum: GypsumRangeV1 | None = None
    Mirrors: MirrorsRangeV1 | None = None
    PaintingAndCoating: PaintingAndCoatingRangeV1 | None = None
    WallFinishes: WallFinishesRangeV1 | None = None
    Plaster: PlasterRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = ("encode_path_param",)

from urllib.parse import quote


def encode_path_param(value: str) -> str:
    """
    Encode a path parameter value.

    :param value: parameter value
    :return: encoded value
    """
    return quote(value, safe="")


def remove_none_id_fields(d: dict) -> dict:
    """
    Remove any key 'id' with a None value from the dictionary, including nested dicts.

    :param d: the dict which may contain 'id' keys with None values.
    :return: a new dict with 'id' keys that have None values removed.

    :note:
        - This function does not modify the original dictionary (no side effects).
        - It returns a new dictionary with the necessary modifications applied.

    :example:
        >>> data = {
        ...     "id": None,
        ...     "name": "item1",
        ...     "details": {
        ...         "id": None,
        ...         "category": "tools",
        ...         "nested": {
        ...             "id": None,
        ...             "value": 42
        ...         }
        ...     }
        ... }
        >>> remove_none_id_fields(data)
        {'name': 'item1', 'details': {'category': 'tools', 'nested': {'value': 42}}}
    """
    if not isinstance(d, dict):
        return d

    cleaned_dict = {}
    for k, v in d.items():
        if isinstance(v, dict):
            v = remove_none_id_fields(v)
        if not (k == "id" and v is None):
            cleaned_dict[k] = v

    return cleaned_dict



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import ClassVar, Literal

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema
from openepd.model.specs.concrete import Cementitious, ConcreteTypicalApplication
from openepd.model.specs.enums import AciExposureClass, CsaExposureClass, EnExposureClass
from openepd.model.specs.singular import BaseCompatibilitySpec
from openepd.model.validation.numbers import RatioFloat
from openepd.model.validation.quantity import LengthInchStr, PressureMPaStr


class ConcreteOptions(BaseOpenEpdSchema):
    """Legacy Concrete options model."""

    lightweight: bool | None = pyd.Field(
        default=None, description="Lightweight. True if < 120lb/ft3 or 1900 kg/m3", example=True
    )
    scc: bool | None = pyd.Field(default=None, description="Self Compacting", example=True)
    finishable: bool | None = pyd.Field(default=None, description="Finishable", example=True)
    air: bool | None = pyd.Field(default=None, description="Air Entrainment", example=True)
    co2_entrain: bool | None = pyd.Field(
        default=None, description="CO2 Curing. Uses intentionally entrained CO2.", example=True
    )
    white_cement: bool | None = pyd.Field(default=None, description="White Cement", example=True)
    plc: bool | None = pyd.Field(default=None, description="Portland Limestone Cement", example=True)
    fiber_reinforced: bool | None = pyd.Field(default=None, description="fiber_reinforced", example=True)


class ConcreteOldSpec(BaseCompatibilitySpec):
    """Old version of the Concrete spec, please use Concrete hierarchical version instead."""

    COMPATIBILITY_SPECS_KEY_OLD: ClassVar[str] = "concrete"
    COMPATIBILITY_SPECS_KEY_NEW: ClassVar[str] = "Concrete"
    COMPATIBILITY_MAPPING: ClassVar[dict[str, str]] = {
        "concrete.strength_28d": "Concrete.strength_28d",
        "concrete.slump": "Concrete.min_slump",
        "concrete.strength_other": "Concrete.strength_other",
        "concrete.strength_other_d": "Concrete.strength_other_d",
        "concrete.w_c_ratio": "Concrete.w_c_ratio",
        "concrete.aci_exposure_classes": "Concrete.aci_exposure_classes",
        "concrete.csa_exposure_classes": "Concrete.csa_exposure_classes",
        "concrete.en_exposure_classes": "Concrete.en_exposure_classes",
        "concrete.application": "Concrete.typical_application",
        "concrete.options.lightweight": "Concrete.lightweight",
        "concrete.options.scc": "Concrete.self_consolidating",
        "concrete.options.finishable": "Concrete.finishable",
        "concrete.options.air": "Concrete.air_entrain",
        "concrete.options.co2_entrain": "Concrete.co2_entrain",
        "concrete.options.white_cement": "Concrete.white_cement",
        "concrete.options.plc": "Concrete.plc",
        "concrete.options.fiber_reinforced": "Concrete.fiber_reinforced",
        "concrete.cementitious": "Concrete.cementitious",
    }

    strength_28d: PressureMPaStr | None = pyd.Field(
        default=None, description="Compressive Strength at 28 days", example="1 MPa"
    )
    slump: LengthInchStr | None = pyd.Field(default=None, description="Minimum test slump", example="2 in")
    strength_other: PressureMPaStr | None = pyd.Field(
        default=None,
        description="One additional strength, which can be early (e.g. 3d) or late (e.g. 96d)",
        example="30 MPa",
    )
    strength_other_d: Literal[3, 7, 14, 42, 56, 72, 96, 120] | None = pyd.Field(
        default=None,
        description="Days for the strength field above. Required IF strength_other is provided.",
        example=42,
    )
    w_c_ratio: RatioFloat | None = pyd.Field(
        default=None, description="Ratio of water to cement", example=0.5, ge=0, le=1
    )
    aci_exposure_classes: list[AciExposureClass] | None = pyd.Field(
        default=None, description="List of ACI318-19 exposure classes this product meets", example=["aci.F0"]
    )
    csa_exposure_classes: list[CsaExposureClass] | None = pyd.Field(
        default=None, description="List of CSA A23.1 exposure classes this product meets", example=["csa.C-2"]
    )
    en_exposure_classes: list[EnExposureClass] | None = pyd.Field(
        default=None, description="List of EN206 exposure classes this product meets", example=["en206.X0"]
    )
    application: ConcreteTypicalApplication | None = pyd.Field(default=None, description="Typical Application")
    options: ConcreteOptions | None = pyd.Field(default=None, description="List of true/false properties")
    cementitious: Cementitious | None = pyd.Field(
        default=None, description="List of cementitious materials, and proportion by mass"
    )



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "LowVoltBusesRangeV1",
    "MedVoltBusesRangeV1",
    "BatteriesRangeV1",
    "OtherElectricalPowerStorageRangeV1",
    "CableTraysRangeV1",
    "ElectricalBusesRangeV1",
    "FloorEquipmentBoxesRangeV1",
    "PowerDistributionUnitsRangeV1",
    "RacewaysRangeV1",
    "FueledElectricalGeneratorsRangeV1",
    "OtherGenerationRangeV1",
    "PhotovoltaicsRangeV1",
    "WindTurbinesRangeV1",
    "ElectricityFromPowerGridRangeV1",
    "ElectricityFromSpecificGeneratorRangeV1",
    "PowerPurchaseAgreementsRangeV1",
    "LightbulbsRangeV1",
    "LightingControlsRangeV1",
    "LightingFixturesRangeV1",
    "OutdoorLightingRangeV1",
    "SpecialtyLightingRangeV1",
    "TaskLightingRangeV1",
    "ElectricalPowerStorageRangeV1",
    "LowVoltageElectricalDistributionRangeV1",
    "ElectricalGenerationEquipmentRangeV1",
    "ElectricPowerRangeV1",
    "LightingRangeV1",
    "ElectricalConduitRangeV1",
    "ElectricalRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.common import RangeFloat
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import CableTraysMaterial, ConduitMaterial, EnergySource, RacewaysMaterial
from openepd.model.validation.quantity import (
    AmountRangeColorTemperature,
    AmountRangeLengthMm,
    AmountRangeLuminosity,
    AmountRangeMass,
    AmountRangePower,
    AmountRangeUtilization,
)


class LowVoltBusesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Busbars and Busways of 600V or less.

    Range version.
    """

    _EXT_VERSION = "1.0"


class MedVoltBusesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Busbars and Busways over 600V.

    Range version.
    """

    _EXT_VERSION = "1.0"


class BatteriesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Battery equipment, including central batteries, battery charging, and UPS.

    Range version.
    """

    _EXT_VERSION = "1.0"


class OtherElectricalPowerStorageRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Other electrical power storage performance specification.

    Range version.
    """

    _EXT_VERSION = "1.0"


class CableTraysRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Mechanical support for electrial or communications cabling, typically suspended from a roof or wall.

    Range version.
    """

    _EXT_VERSION = "1.0"

    height: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    width: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    depth: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    static_load: AmountRangeMass | None = pyd.Field(default=None, description="")
    ventilated: bool | None = pyd.Field(default=None, description="At least 40% of the tray base is open to air flow")
    cable_trays_material: list[CableTraysMaterial] | None = pyd.Field(default=None, description="")


class ElectricalBusesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Power distribution, in the form of busbars or of insulted ducts made of copper or aluminum busbars.

    It is an alternative means of conducting electricity compared toto power cables or cable bus. Also called
    bus ducts.

    Range version.
    """

    _EXT_VERSION = "1.0"

    LowVoltBuses: LowVoltBusesRangeV1 | None = None
    MedVoltBuses: MedVoltBusesRangeV1 | None = None


class FloorEquipmentBoxesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Equipment boxes for power or electronic equipment embedded in an accessible floor.

    Range version.
    """

    _EXT_VERSION = "1.0"


class PowerDistributionUnitsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Switched electrical distribution units placed very close to the point of consumption, for example inside a rack of electronic equipment.

    Range version.
    """

    _EXT_VERSION = "1.0"


class RacewaysRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Mechanical guideways for eletrical communications cabling, typically embedded in an accessible floor.

    Range version.
    """

    _EXT_VERSION = "1.0"

    width: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    depth: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    painted: bool | None = pyd.Field(default=None, description="")
    divided: bool | None = pyd.Field(default=None, description="")
    raceways_material: list[RacewaysMaterial] | None = pyd.Field(default=None, description="")


class FueledElectricalGeneratorsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Fueled electrical generators.

    Range version.
    """

    _EXT_VERSION = "1.0"


class OtherGenerationRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Other generation.

    Range version.
    """

    _EXT_VERSION = "1.0"


class PhotovoltaicsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Solar photovoltaics, rated on a nameplate capacity basis.

    Range version.
    """

    _EXT_VERSION = "1.0"


class WindTurbinesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Wind generators, rated on a nameplate capacity basis.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ElectricityFromPowerGridRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Electrical energy drawn from a specific utility grid.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ElectricityFromSpecificGeneratorRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Electrical energy from a specific power plant, such as a wind farm using a specific type of turbine.

    Range version.
    """

    _EXT_VERSION = "1.0"

    energy_source: list[EnergySource] | None = pyd.Field(default=None, description="")


class PowerPurchaseAgreementsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Electrical energy subject to a verified power purchase agreement.

    The impact of electricity generation is allocated specifically to the agreement and not to the general grid.

    Range version.
    """

    _EXT_VERSION = "1.0"

    energy_source: list[EnergySource] | None = pyd.Field(default=None, description="")


class LightbulbsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Various types of light bulbs, including LED, CFL, halogen, and incandescent.

    Range version.
    """

    _EXT_VERSION = "1.0"


class LightingControlsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Devices used to control the operation of lighting, including dimmers, sensors, and smart controls.

    Range version.
    """

    _EXT_VERSION = "1.0"


class LightingFixturesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Permanent lighting fixtures for interior spaces, including ceiling, wall-mounted, and pendant fixtures.

    Range version.
    """

    _EXT_VERSION = "1.0"


class OutdoorLightingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Lighting products designed for outdoor use, including landscape and security lighting.

    Range version.
    """

    _EXT_VERSION = "1.0"


class SpecialtyLightingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Specialized lighting for niche applications like emergency, medical, or theatrical lighting.

    Range version.
    """

    _EXT_VERSION = "1.0"


class TaskLightingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Lighting designed for specific tasks such as desk lamps, under-cabinet lighting, and reading lamps.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ElectricalPowerStorageRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Electrical Power Storage.

    Range version.
    """

    _EXT_VERSION = "1.0"

    Batteries: BatteriesRangeV1 | None = None
    OtherElectricalPowerStorage: OtherElectricalPowerStorageRangeV1 | None = None


class LowVoltageElectricalDistributionRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Low Voltage Electrical Distribution.

    Range version.
    """

    _EXT_VERSION = "1.0"

    CableTrays: CableTraysRangeV1 | None = None
    ElectricalBuses: ElectricalBusesRangeV1 | None = None
    FloorEquipmentBoxes: FloorEquipmentBoxesRangeV1 | None = None
    PowerDistributionUnits: PowerDistributionUnitsRangeV1 | None = None
    Raceways: RacewaysRangeV1 | None = None


class ElectricalGenerationEquipmentRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Equipment for generating electrical power.

    This category is primarily for smaller-scale. (e.g. on premises) generation, rather than utility-scale equipment.

    Range version.
    """

    _EXT_VERSION = "1.0"

    FueledElectricalGenerators: FueledElectricalGeneratorsRangeV1 | None = None
    OtherGeneration: OtherGenerationRangeV1 | None = None
    Photovoltaics: PhotovoltaicsRangeV1 | None = None
    WindTurbines: WindTurbinesRangeV1 | None = None


class ElectricPowerRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Electrical energy drawn from a utility grid.

    Range version.
    """

    _EXT_VERSION = "1.0"

    ElectricityFromPowerGrid: ElectricityFromPowerGridRangeV1 | None = None
    ElectricityFromSpecificGenerator: ElectricityFromSpecificGeneratorRangeV1 | None = None
    PowerPurchaseAgreements: PowerPurchaseAgreementsRangeV1 | None = None


class LightingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Lamps and lightbulbs and lamp components.

    Range version.
    """

    _EXT_VERSION = "1.0"

    color_temperature: AmountRangeColorTemperature | None = pyd.Field(default=None, description="")
    typical_utilization: AmountRangeUtilization | None = pyd.Field(default=None, description="")
    luminosity: AmountRangeLuminosity | None = pyd.Field(default=None, description="")
    wattage: AmountRangePower | None = pyd.Field(default=None, description="")
    color_rendering_index: RangeFloat | None = pyd.Field(default=None, description="")
    dimmable: bool | None = pyd.Field(default=None, description="")
    Lightbulbs: LightbulbsRangeV1 | None = None
    LightingControls: LightingControlsRangeV1 | None = None
    LightingFixtures: LightingFixturesRangeV1 | None = None
    OutdoorLighting: OutdoorLightingRangeV1 | None = None
    SpecialtyLighting: SpecialtyLightingRangeV1 | None = None
    TaskLighting: TaskLightingRangeV1 | None = None


class ElectricalConduitRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Tubing used to protect and route electrical wiring in a building or structure.

    Range version.
    """

    _EXT_VERSION = "1.1"

    nominal_diameter: AmountRangeLengthMm | None = None
    outer_diameter: AmountRangeLengthMm | None = None
    inner_diameter: AmountRangeLengthMm | None = None
    wall_thickness: AmountRangeLengthMm | None = None
    material: list[ConduitMaterial] | None = None


class ElectricalRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Electric power and equipment.

    Range version.
    """

    _EXT_VERSION = "1.1"

    ElectricalPowerStorage: ElectricalPowerStorageRangeV1 | None = None
    LowVoltageElectricalDistribution: LowVoltageElectricalDistributionRangeV1 | None = None
    ElectricalGenerationEquipment: ElectricalGenerationEquipmentRangeV1 | None = None
    ElectricPower: ElectricPowerRangeV1 | None = None
    Lighting: LightingRangeV1 | None = None
    ElectricalConduit: ElectricalConduitRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import datetime
from enum import StrEnum
from typing import Annotated, Optional

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema, OpenXpdUUID
from openepd.model.common import Amount, WithAltIdsMixin, WithAttachmentsMixin
from openepd.model.org import Org


class PcrStatus(StrEnum):
    """Status of a PCR."""

    InDevelopment = "InDevelopment"
    Published = "Published"
    NonPublic = "NonPublic"
    Expired = "Expired"
    Sunset = "Sunset"


class PcrRef(BaseOpenEpdSchema):
    """Reference to a PCR."""

    id: OpenXpdUUID | None = pyd.Field(
        description="The unique ID for this PCR.  To ensure global uniqueness, should be registered "
        "at open-xpd-uuid.cqd.io/register or a coordinating registry.",
        example="ec3xpgq2",
        default=None,
    )
    name: str | None = pyd.Field(
        max_length=200,
        description="Full document name as listed in source document",
        example="c-PCR-003 Concrete and concrete elements (EN 16757)",
    )
    ref: pyd.AnyUrl | None = pyd.Field(
        description="Reference to this PCR's JSON object",
        example="https://openepd.buildingtransparency.org/api/pcrs/1u7zsed8",
    )


class Pcr(WithAttachmentsMixin, WithAltIdsMixin, BaseOpenEpdSchema):
    """Represent a PCR (Product Category Rules)."""

    id: OpenXpdUUID | None = pyd.Field(
        description="The unique ID for this PCR.  To ensure global uniqueness, should be registered "
        "at open-xpd-uuid.cqd.io/register or a coordinating registry.",
        example="ec3xpgq2",
        default=None,
    )
    issuer: Org | None = pyd.Field(description="Organization issuing this PCR", default=None)
    issuer_doc_id: str | None = pyd.Field(
        max_length=40,
        default=None,
        description="Document ID or code created by issuer",
        example="c-PCR-003",
    )
    name: str | None = pyd.Field(
        max_length=200,
        default=None,
        description="Full document name as listed in source document",
        example="c-PCR-003 Concrete and concrete elements (EN 16757)",
    )
    short_name: str | None = pyd.Field(
        default=None,
        description="A shortened name without boilerplate text.",
        example="Concrete and Concrete Elements",
    )
    declared_units: list[Amount] | None = pyd.Field(
        description="SI declared units for this PCR.  If a functional unit is "
        "utilized, the declared unit shall refer to the amount of "
        "product associated with the A1-A3 life cycle stage."
    )
    version: str | None = pyd.Field(
        description="Document version, as expressed in document.",
        example="1.0.2",
        default=None,
    )
    date_of_issue: datetime.datetime | None = pyd.Field(
        example=datetime.datetime(day=11, month=9, year=2019, tzinfo=datetime.timezone.utc),
        default=None,
        description="First day on which the document is valid",
    )
    valid_until: datetime.datetime | None = pyd.Field(
        example=datetime.datetime(day=11, month=9, year=2019, tzinfo=datetime.timezone.utc),
        default=None,
        description="Last day on which the document is valid",
    )
    doc: str | None = pyd.Field(default=None, description="URL to original document, preferably directly to a PDF.")
    parent: Optional["Pcr"] = pyd.Field(
        description="The parent PCR, base PCR, `Part A` PCR",
        default=None,
    )
    status: PcrStatus | None = pyd.Field(
        default=None,
        description="The current release status of this PCR. "
        "A PCR with valid_until in the past must have status Expired or Sunset; a PCR with valid_until "
        "more than 5 years in the past must have status Sunset. Compliant systems should automatically "
        "update these fields within 24 hours.",
    )
    product_classes: dict[str, str | list[str]] = pyd.Field(
        description="List of classifications, including Masterformat and UNSPC", default_factory=dict
    )
    applicable_in: list[Annotated[str, pyd.Field(min_length=2, max_length=2)]] | None = pyd.Field(
        max_items=100,
        default=None,
        description="Jurisdiction(s) in which EPD is applicable. An empty array, or absent properties, "
        "implies global applicability. Accepts "
        "[2-letter country codes](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2), "
        "[M49 region codes](https://unstats.un.org/unsd/methodology/m49/), "
        'or the alias "EU27" for the 27 members of the Euro bloc, or the alias "NAFTA" '
        "for the members of North American Free Trade Agreement",
        example=["US", "CA", "MX", "EU27", "NAFTA"],
    )

    @classmethod
    def get_asset_type(cls) -> str | None:
        """Return the asset type of this class (see BaseOpenEpdSchema.get_asset_type for details)."""
        return "pcr"



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import abc
import datetime
from enum import StrEnum

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema, OpenXpdUUID, RootDocument
from openepd.model.common import Amount
from openepd.model.geography import Geography
from openepd.model.org import Org
from openepd.model.pcr import Pcr
from openepd.model.specs.range import SpecsRange
from openepd.model.standard import Standard
from openepd.model.validation.common import ReferenceStr
from openepd.model.validation.quantity import AmountGWP, AmountMass

DEVELOPER_DESCRIPTION = "The organization responsible for the underlying LCA (and subsequent summarization as EPD)."
PROGRAM_OPERATOR_DESCRIPTION = "JSON object for program operator Org"
THIRD_PARTY_VERIFIER_DESCRIPTION = "JSON object for Org that performed a critical review of the EPD data"


class BaseDeclaration(RootDocument, abc.ABC):
    """Base class for declaration-related documents (EPDs, Industry-wide EPDs, Generic Estimates)."""

    id: OpenXpdUUID | None = pyd.Field(
        description="The unique ID for this document.  To ensure global uniqueness, should be registered at "
        "open-xpd-uuid.cqd.io/register or a coordinating registry.",
        example="1u7zsed8",
        default=None,
    )
    date_of_issue: datetime.datetime | None = pyd.Field(
        example=datetime.datetime(day=11, month=9, year=2019, tzinfo=datetime.timezone.utc),
        description="Date the document was issued. This should be the first day on which the document is valid.",
    )
    valid_until: datetime.datetime | None = pyd.Field(
        example=datetime.datetime(day=11, month=9, year=2028, tzinfo=datetime.timezone.utc),
        description="Last date the document is valid on, including any extensions.",
    )

    version: pyd.NonNegativeInt | None = pyd.Field(
        description="Version of this document. The document's issuer should increment it anytime even a single "
        "character changes, as this value is used to determine the most recent version.",
        example=1,
        default=None,
    )

    declared_unit: Amount | None = pyd.Field(
        description="SI declared unit for this document.  If a functional unit is "
        "utilized, the declared unit shall refer to the amount of "
        "product associated with the A1-A3 life cycle stage."
    )
    kg_per_declared_unit: AmountMass | None = pyd.Field(
        default=None,
        description="Mass of the product, in kilograms, per declared unit",
        example=Amount(qty=12.5, unit="kg").to_serializable(exclude_unset=True),
    )
    compliance: list[Standard] = pyd.Field(
        description="Standard(s) to which this document is compliant.", default_factory=list
    )

    # TODO: add product_alt_names? E.g. ILCD has a list of synonymous names
    product_classes: dict[str, str | list[str]] = pyd.Field(
        description="List of classifications, including Masterformat and UNSPC", default_factory=dict
    )

    language: str | None = pyd.Field(
        min_length=2,
        max_length=2,
        strip_whitespace=True,
        description="Language this document is captured in, as an ISO 639-1 code",
        example="en",
    )
    private: bool | None = pyd.Field(
        default=False,
        description="This document's author does not wish the contents published. "
        "Useful for draft, partial, or confidential declarations.  "
        "How (or whether) privacy is implemented is up to the receiving system.  "
        "Null is treated as false (public).  Private (draft) entries have a reduced "
        "number of required fields, to allow for multiple systems to coordinate "
        "incomplete EPDs.",
    )

    pcr: Pcr | None = pyd.Field(
        description="JSON object for product category rules. Should point to the "
        "most-specific PCR that applies; the PCR entry should point to any "
        "parent PCR.",
        default=None,
    )
    lca_discussion: str | None = pyd.Field(
        max_length=20000,
        description="""A rich text description containing information for experts reviewing the document contents.
    Text descriptions required by ISO 14025, ISO 21930, EN 15804,, relevant PCRs, or program instructions and which do not
    have specific openEPD fields should be entered here.  This field may be large, and may contain multiple sections
    separated by github flavored markdown formatting.""",
        example="""# Packaging

    Information on product-specific packaging: type, composition and possible reuse of packaging materials (paper,
    strapping, pallets, foils, drums, etc.) shall be included in this Section. The EPD shall describe specific packaging
    scenario assumptions, including disposition pathways for each packaging material by reuse, recycling, or landfill
    disposal based on packaging type.*

    # Product Installation

    A description of the type of processing, machinery, tools, dust extraction equipment, auxiliary materials, etc.
    to be used during installation shall be included. Information on industrial and environmental protection may be
    included in this section. Any waste treatment included within the system boundary of installation waste should be
    specified.

    # Use Conditions

    Use-stage environmental impacts of flooring products during building operations depend on product cleaning assumptions.
    Information on cleaning frequency and cleaning products shall be provided based on the manufacturer’s recommendations.
    In the absence of primary data, cleaning assumptions shall be documented.
    """,
    )

    product_image_small: pyd.AnyUrl | None = pyd.Field(
        description="Pointer to image illustrating the product, which is no more than 200x200 pixels", default=None
    )
    product_image: pyd.AnyUrl | pyd.FileUrl | None = pyd.Field(
        description="pointer to image illustrating the product no more than 10MB", default=None
    )
    declaration_url: str | None = pyd.Field(
        description="Link to data object on original registrar's site",
        example="https://epd-online.com/EmbeddedEpdList/Download/6029",
    )
    kg_C_per_declared_unit: AmountMass | None = pyd.Field(
        default=None,
        description="Mass of elemental carbon, per declared unit, contained in the product itself at the manufacturing "
        "facility gate.  Used (among other things) to check a carbon balance or calculate incineration "
        "emissions.  The source of carbon (e.g. biogenic) is not relevant in this field.",
        example=Amount(qty=8.76, unit="kgCO2e").to_serializable(exclude_unset=True),
    )
    kg_C_biogenic_per_declared_unit: AmountGWP | None = pyd.Field(
        default=None,
        description="Mass of elemental carbon from biogenic sources, per declared unit, contained in the product "
        "itself at the manufacturing facility gate.  It may be presumed that any biogenic carbon content "
        "has been accounted for as -44/12 kgCO2e per kg C in stages A1-A3, per EN15804 and ISO 21930.",
        example=Amount(qty=8.76, unit="kgCO2e").to_serializable(exclude_unset=True),
    )
    product_service_life_years: float | None = pyd.Field(
        gt=0.0009,
        lt=101,
        description="Reference service life of the product, in years.  Serves as a maximum for replacement interval, "
        "which may also be constrained by usage or the service life of what the product goes into "
        "(e.g. a building).",
        example=50.0,
    )


class AverageDatasetMixin(pyd.BaseModel, title="Average Dataset"):
    """Fields common for average dataset (Industry-wide EPDs, Generic Estimates)."""

    description: str | None = pyd.Field(
        max_length=2000,
        description="1-paragraph description of the average dataset. Supports plain text or github flavored markdown.",
    )

    geography: list[Geography] | None = pyd.Field(
        description="Jurisdiction(s) in which the LCA result is applicable.  An empty array, or absent properties, "
        "implies global applicability.",
    )

    specs: SpecsRange | None = pyd.Field(
        default=None, description="Average dataset material performance specifications."
    )


class WithProgramOperatorMixin(pyd.BaseModel):
    """Object which has a connection to ProgramOperator."""

    program_operator: Org | None = pyd.Field(description=PROGRAM_OPERATOR_DESCRIPTION)
    program_operator_doc_id: str | None = pyd.Field(
        max_length=200, description="Document identifier from Program Operator.", example="123-456.789/b"
    )
    program_operator_version: str | None = pyd.Field(
        max_length=200, description="Document version number from Program Operator.", example="4.3.0"
    )


class WithVerifierMixin(pyd.BaseModel):
    """Set of fields related to verifier."""

    third_party_verifier: Org | None = pyd.Field(description=THIRD_PARTY_VERIFIER_DESCRIPTION)
    third_party_verification_url: pyd.AnyUrl | None = pyd.Field(
        description="Optional link to a verification statement.",
        example="https://we-verify-epds.com/en/letters/123-456.789b.pdf",
    )
    third_party_verifier_email: pyd.EmailStr | None = pyd.Field(
        description="Email address of the third party verifier", example="john.doe@example.com", default=None
    )


class WithEpdDeveloperMixin(pyd.BaseModel):
    """Set of fields related to EPD Developer."""

    epd_developer: Org | None = pyd.Field(
        description=DEVELOPER_DESCRIPTION,
        default=None,
    )
    epd_developer_email: pyd.EmailStr | None = pyd.Field(
        default=None,
        example="john.doe@we-do-lca.com",
        description="Email contact for inquiries about development of this EPD. "
        "This must be an email which can be publicly shared.",
    )


class RefBase(BaseOpenEpdSchema, title="Ref Object"):
    """Base class for reference-style objects."""

    id: OpenXpdUUID | None = pyd.Field(
        description="The unique ID for this object. To ensure global uniqueness, should be registered at "
        "open-xpd-uuid.cqd.io/register or a coordinating registry.",
        example="1u7zsed8",
        default=None,
    )

    name: str | None = pyd.Field(max_length=200, description="Name of the object", default=None)

    ref: ReferenceStr | None = pyd.Field(
        default=None,
        example="https://openepd.buildingtransparency.org/api/generic_estimates/EC300001",
        description="Reference to this JSON object",
    )


class OriginalDataFormat(StrEnum):
    """
    Original data format for this EPD.

    A system receiving an EPD via openEPD should preserve this field if it exists.  Otherwise, it is up to the
    receiving system to infer the original data format based on the source.
    """

    OPENEPD_1_0 = "openEPDv1.0"
    """
    Data was generated in an openEPD-compliant system, such as by manual entry in EC3 or generation in an openEPD
    compliant EPD generator.
    """

    CUSTOM_API = "custom_api"
    """
    Data was gathered from a non-standardized API offered by a data provider.  Numerical and text data is very likely
    to be correct and up to date, but important fields may be absent and identification of organizations, plants, and
     addresses may be ambiguous.
    """

    ILCD_EPD_v1_2 = "ILCD+EPDv1_2"
    """
    Data was generated and published in ILCD+EPD format, version 1.2  Numerical and text data is very likely to be
    correct and up to date, but important fields may be absent and identification of organizations, plants, and
    addresses may be ambiguous.
    """

    ILCD_EPD_v1_1 = "ILCD+EPDv1_1"
    """
    Data was generated and published in ILCD+EPD format, version 1.1  Numerical and text data is very likely to be
    correct and up to date, but important fields may be absent and identification of organizations, plants, and
    addresses may be ambiguous.
    """

    ILCD_EPD_v1_0 = "ILCD+EPDv1_0"
    """
    Data was generated and published in ILCD+EPD format, version 1.0  Numerical and text data is very likely to be
    correct and up to date, but important fields may be absent and identification of organizations, plants, and
    addresses may be ambiguous.
    """

    WEB = "web"
    """
    Data was gathered through comprehension of a viewable HTML page, typically offered by a program operator.
    This is more reliable than PDF, but less reliable than API or a standardized XML like ILCD+EPD.
    """

    PDF_MANUAL = "pdf_manual"
    """
    Data was entered by hand based on PDF and/or validated by a qualified professional.
    """

    PDF_MIXED = "pdf_mixed"
    """
    Data is a mix of structured data from an information system, but with significant missing data filled in by
    extraction from a PDF.
    """

    PDF = "pdf"
    """
    Data was extracted by analyzing or comprehending a PDF document.  Includes documents where up to 4 fields (e.g.
    program operator) were inferred from sources outside the PDF.  This method is prone to errors and omissions due
    to the many different formats and terms used in these relatively unstructured documents.
    """

    OTHER = "other"
    """
    Data source is not of a type listed here.
    """



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = ("WoodJoistsRangeV1",)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.org import OrgRef
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import AllFabrication, AllTimberSpecies


class WoodJoistsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Wood joists.

    Prefabricated I-shaped engineered wood structural members made primarily from one or more types of wood. Includes
    products made with metallic webbing. Excludes products where the wood is merely decorative.

    Range version.
    """

    _EXT_VERSION = "1.0"

    forest_practices_certifiers: list[OrgRef] | None = None
    timber_species: list[AllTimberSpecies] | None = pyd.Field(default=None, description="")
    fabrication: list[AllFabrication] | None = pyd.Field(default=None, description="")
    weather_exposed: bool | None = pyd.Field(default=None, description="")
    fire_retardant: bool | None = pyd.Field(default=None, description="")
    decay_resistant: bool | None = pyd.Field(default=None, description="")



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "ArchitecturalPrecastRangeV1",
    "StructuralPrecastRangeV1",
    "UtilityUndergroundPrecastRangeV1",
    "CivilPrecastRangeV1",
    "PrecastConcreteRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.common import RangeRatioFloat
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.singular.precast_concrete import (
    ArchitecturalPrecastElementType,
    CivilPrecastElementType,
    StructuralPrecastElementType,
    UtilityPrecastElementType,
)
from openepd.model.validation.quantity import AmountRangePressureMpa


class ArchitecturalPrecastRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Precast concrete cladding used for architectural purposes.

    Range version.
    """

    _EXT_VERSION = "1.1"

    element_type: ArchitecturalPrecastElementType | None = pyd.Field(
        default=None, description="Precast element type used for architectural applications."
    )


class StructuralPrecastRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Precast concrete used for structural purposes.

    Range version.
    """

    _EXT_VERSION = "1.1"

    element_type: StructuralPrecastElementType | None = pyd.Field(
        default=None, description="Precast element type used for structural applications."
    )


class UtilityUndergroundPrecastRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Precast concrete for utility vaults, manhole, drain inlets. Excludes piping.

    Range version.
    """

    _EXT_VERSION = "1.1"

    element_type: UtilityPrecastElementType | None = pyd.Field(
        default=None, description="Precast element type used for utility underground applications."
    )


class CivilPrecastRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Precast concrete used for civil engineering applications including bridges, highways, and railroads.

    Range version.
    """

    _EXT_VERSION = "1.0"

    element_type: CivilPrecastElementType | None = pyd.Field(
        default=None, description="Precast element type used as civil engineering components."
    )


class PrecastConcreteRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    General category for precast concrete components.

    Range version.
    """

    _EXT_VERSION = "1.1"

    lightweight: bool | None = pyd.Field(default=None, description="")
    concrete_compressive_strength_28d: AmountRangePressureMpa | None = pyd.Field(default=None, description="")
    insulated: bool | None = pyd.Field(default=None, description="")
    gfrc: bool | None = pyd.Field(
        default=None,
        description="Glass Fiber Reinforced Concrete is fiber-reinforced concrete sometimes used in architectural panels",
    )
    steel_mass_percentage: RangeRatioFloat | None = pyd.Field(default=None, description="")
    ArchitecturalPrecast: ArchitecturalPrecastRangeV1 | None = None
    StructuralPrecast: StructuralPrecastRangeV1 | None = None
    UtilityUndergroundPrecast: UtilityUndergroundPrecastRangeV1 | None = None
    CivilPrecast: CivilPrecastRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from openepd.model.specs.singular.accessories import AccessoriesV1
from openepd.model.specs.singular.aggregates import AggregatesV1
from openepd.model.specs.singular.aluminium import AluminiumV1
from openepd.model.specs.singular.asphalt import AsphaltV1
from openepd.model.specs.singular.bulk_materials import BulkMaterialsV1
from openepd.model.specs.singular.cast_decks_and_underlayment import CastDecksAndUnderlaymentV1
from openepd.model.specs.singular.cladding import CladdingV1
from openepd.model.specs.singular.cmu import CMUV1
from openepd.model.specs.singular.concrete import ConcreteV1
from openepd.model.specs.singular.conveying_equipment import ConveyingEquipmentV1
from openepd.model.specs.singular.electrical import ElectricalV1
from openepd.model.specs.singular.electrical_transmission_and_distribution_equipment import (
    ElectricalTransmissionAndDistributionEquipmentV1,
)
from openepd.model.specs.singular.electricity import ElectricityV1
from openepd.model.specs.singular.finishes import FinishesV1
from openepd.model.specs.singular.fire_and_smoke_protection import FireAndSmokeProtectionV1
from openepd.model.specs.singular.furnishings import FurnishingsV1
from openepd.model.specs.singular.grouting import GroutingV1
from openepd.model.specs.singular.manufacturing_inputs import ManufacturingInputsV1
from openepd.model.specs.singular.masonry import MasonryV1
from openepd.model.specs.singular.material_handling import MaterialHandlingV1
from openepd.model.specs.singular.mechanical import MechanicalV1
from openepd.model.specs.singular.mechanical_insulation import MechanicalInsulationV1
from openepd.model.specs.singular.network_infrastructure import NetworkInfrastructureV1
from openepd.model.specs.singular.openings import OpeningsV1
from openepd.model.specs.singular.other_electrical_equipment import OtherElectricalEquipmentV1
from openepd.model.specs.singular.other_materials import OtherMaterialsV1
from openepd.model.specs.singular.plumbing import PlumbingV1
from openepd.model.specs.singular.precast_concrete import PrecastConcreteV1
from openepd.model.specs.singular.sheathing import SheathingV1
from openepd.model.specs.singular.steel import SteelV1
from openepd.model.specs.singular.thermal_moisture_protection import ThermalMoistureProtectionV1
from openepd.model.specs.singular.utility_piping import UtilityPipingV1
from openepd.model.specs.singular.wood import WoodV1
from openepd.model.specs.singular.wood_joists import WoodJoistsV1



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, ClassVar

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema
from openepd.model.common import Amount, OpenEPDUnit, RangeAmount

if TYPE_CHECKING:
    QuantityValidatorType = Callable[[type, str], str]


class QuantityValidator(ABC):
    """
    Interface for quantity validator.

    The openEPD models are mapped using the simple types. Caller code should provide their own implementation of this
    and set it with `set_unit_validator` function.
    """

    @abstractmethod
    def validate_same_dimensionality(self, unit: str | None, dimensionality_unit: str | None) -> None:
        """
        Validate that a given unit ('kg') has the same dimesnionality as provided dimensionality_unit ('g').

        :param unit: unit to validate, not quantity
        :param dimensionality_unit: unit to check against
        :raise:
            ValueError if dimensionality is different
        :return: None
        """
        pass

    @abstractmethod
    def validate_unit_correctness(self, value: str, dimensionality: str) -> None:
        """
        Validate the given string value against the given dimensionality.

        Args:
            value: The value to validate, like "102.4 kg"
            dimensionality: The dimensionality to validate against, like "kg"
        Returns:
            None if the value is valid, raises an error otherwise.
        Raises:
            ValueError: If the value is not valid.
        """
        pass

    @abstractmethod
    def validate_quantity_greater_or_equal(self, value: str, min_value: str) -> None:
        """
        Validate the quantity is greater than minimal value.

        Args:
            value: The value to validate, like "2.4 kg"
            min_value: The value to compare with, like "102.4 kg"
        Returns:
            None if the value is valid, raises an error otherwise.
        Raises:
            ValueError: If the value is not valid.
        """
        pass

    @abstractmethod
    def validate_quantity_less_or_equal(self, value: str, max_value: str) -> None:
        """
        Validate the quantity is less than minimal value.

        Args:
            value: The value to validate, like "2.4 kg"
            max_value: The value to compare with, like "0.4 kg"
        Returns:
            None if the value is valid, raises an error otherwise.
        Raises:
            ValueError: If the value is not valid.
        """
        pass


class ExternalValidationConfig:
    """
    Configuration holder for external validator.

    Since openEPD library does not provide any facility for working with quantities/units, users should do this
    by implementing the protocol for this validator and setting it with setup_external_validators function.
    """

    QUANTITY_VALIDATOR: ClassVar[QuantityValidator | None] = None


def validate_unit_factory(dimensionality: OpenEPDUnit | str) -> "QuantityValidatorType":
    """Create validator for units (not quantities) to check for dimensionality."""

    def validator(cls: type | None, value: str) -> str:
        if ExternalValidationConfig.QUANTITY_VALIDATOR is not None:
            ExternalValidationConfig.QUANTITY_VALIDATOR.validate_same_dimensionality(value, dimensionality)
        return value

    return validator


def validate_quantity_unit_factory(dimensionality: OpenEPDUnit | str) -> "QuantityValidatorType":
    """Create validator for quantity field to check unit matching."""

    def validator(cls: type | None, value: str) -> str:
        if ExternalValidationConfig.QUANTITY_VALIDATOR is not None:
            ExternalValidationConfig.QUANTITY_VALIDATOR.validate_unit_correctness(value, dimensionality)
        return value

    return validator


def validate_quantity_ge_factory(min_value: str) -> "QuantityValidatorType":
    """Create validator to check that quantity is greater than or equal to min_value."""

    def validator(cls: type | None, value: str) -> str:
        if ExternalValidationConfig.QUANTITY_VALIDATOR is not None:
            ExternalValidationConfig.QUANTITY_VALIDATOR.validate_quantity_greater_or_equal(value, min_value)
        return value

    return validator


def validate_quantity_le_factory(max_value: str) -> "QuantityValidatorType":
    """Create validator to check that quantity is less than or equal to max_value."""

    def validator(cls: type | None, value: str) -> str:
        if ExternalValidationConfig.QUANTITY_VALIDATOR is not None:
            ExternalValidationConfig.QUANTITY_VALIDATOR.validate_quantity_less_or_equal(value, max_value)
        return value

    return validator


# for arbitrary non-standard quantity
# todo these types should be replaced by Annotated[str, AfterValidator...] as we move completely to pydantic 2


class QuantityStr(str):
    """
    Quantity string type.

    Should be used in models where the physical value (quantity) is expected.

    Checks for dimensionality and for the fact that value is greater than zero.
    """

    unit: ClassVar[str]

    @classmethod
    def __get_validators__(cls):
        unit = getattr(cls, "unit", None)
        if unit:
            yield validate_quantity_unit_factory(cls.unit)
            yield validate_quantity_ge_factory(f"0 {cls.unit}")

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
            example=f"1 {cls.unit}",
        )


class PressureMPaStr(QuantityStr):
    """Pressure quantity type."""

    unit = OpenEPDUnit.MPa


class MassKgStr(QuantityStr):
    """Mass quantity type."""

    unit = OpenEPDUnit.kg


class AreaM2Str(QuantityStr):
    """Area quantity type."""

    unit = OpenEPDUnit.m2


class LengthMStr(QuantityStr):
    """Length (m) quantity type."""

    unit = OpenEPDUnit.m


class LengthMmStr(QuantityStr):
    """Length (mm) quantity type."""

    unit = OpenEPDUnit.m

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
            example="6 mm",
        )


class LengthInchStr(QuantityStr):
    """Length (inch) quantity type."""

    unit = "inch"

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
            example="2.5 inch",
        )


class TemperatureCStr(QuantityStr):
    """Temperature celsius quantity type."""

    unit = OpenEPDUnit.degree_c


class GwpKgCo2eStr(QuantityStr):
    """GWP intensity quantity type."""

    unit = OpenEPDUnit.kg_co2


class CapacityPerHourStr(QuantityStr):
    """Capacity per-hour quantity type."""

    unit = f"{OpenEPDUnit.hour}^-1"


class RValueStr(QuantityStr):
    """R-Value quantity type."""

    unit = "K * m2 / W"


class SpeedStr(QuantityStr):
    """Speed quantity type."""

    unit = "m / s"


class ColorTemperatureStr(QuantityStr):
    """Color temp quantity type."""

    unit = "K"


class LuminosityStr(QuantityStr):
    """Luminosity quantity type."""

    unit = "lumen"


class PowerStr(QuantityStr):
    """Power quantity type."""

    unit = "W"


class ElectricalCurrentStr(QuantityStr):
    """Current quantity type."""

    unit = "A"


class VolumeStr(QuantityStr):
    """Volume quantity type."""

    unit = "m3"


class AirflowStr(QuantityStr):
    """Air flow quantity type."""

    unit = "m3 / s"


class FlowRateStr(QuantityStr):
    """Liquid flow rate quantity type."""

    unit = "l / min"


class MassPerLengthStr(QuantityStr):
    """Mass per unit of length quantity type."""

    unit = "kg / m"


class AreaPerVolumeStr(QuantityStr):
    """Area per unit of volume quantity type."""

    unit = "m2 / l"


class ThermalConductivityStr(QuantityStr):
    """Thermal conductivity quantity type."""

    unit = "1 W / K / m"


class ThermalExpansionStr(QuantityStr):
    """Thermal conductivity quantity type."""

    unit = "1 / K"


class ForceNStr(QuantityStr):
    """Force (N) type."""

    unit = "1 N"


class YarnWeightStr(QuantityStr):
    """Yarn weight quantity type."""

    unit = "1 g / m2"


class UtilizationStr(QuantityStr):
    """Utilization quantity type."""

    unit = "1 h / yr"


class UFactorStr(QuantityStr):
    """Utilization quantity type."""

    unit = "1 USI"


class RFactorStr(QuantityStr):
    """R-Factor type."""

    unit = "1 RSI"


class WithDimensionalityMixin(BaseOpenEpdSchema):
    """Class for dimensionality-validated amounts."""

    dimensionality_unit: ClassVar[str | None] = None

    # Unit for dimensionality to validate against, for example "kg"

    @pyd.root_validator
    def check_dimensionality_matches(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Check that this amount conforms to the same dimensionality as dimensionality_unit."""
        if not cls.dimensionality_unit:
            return values

        validate_unit_factory(cls.dimensionality_unit)(BaseOpenEpdSchema, values.get("unit"))  # type:ignore [arg-type]
        return values


class AmountRangeWithDimensionality(RangeAmount, WithDimensionalityMixin):
    """Mass amount, range."""

    class Config:
        """Pydantic config."""

        @staticmethod
        def schema_extra(schema: dict[str, Any], model: type["AmountRangeWithDimensionality"]) -> None:
            """Modify json schema."""
            schema["example"] = {"min": 1.2, "max": 3.4, "unit": str(model.dimensionality_unit) or None}


class WithMassKgMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = MassKgStr.unit


class AmountMass(Amount, WithMassKgMixin):
    """Amount of mass, measured in kg, t, etc."""

    pass


class AmountRangeMass(AmountRangeWithDimensionality, WithMassKgMixin):
    """Range of masses."""

    pass


class WithGwpMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = GwpKgCo2eStr.unit


class AmountGWP(Amount, WithGwpMixin):
    """Amount of Global Warming Potential, measured in kgCO2e."""

    pass


class AmountRangeGWP(AmountRangeWithDimensionality, WithGwpMixin):
    """Range of masses."""

    pass


class WithLengthMMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    pass


class AmountRangeLengthM(AmountRangeWithDimensionality, WithLengthMMixin):
    """Range of lengths (m)."""

    pass


class AmountLengthM(Amount, WithLengthMMixin):
    """Length (m)."""

    pass


class WithLengthMmMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = LengthMmStr.unit


class AmountLengthMm(Amount, WithLengthMMixin):
    """Length (mm)."""

    pass


class AmountRangeLengthMm(AmountRangeWithDimensionality, WithLengthMmMixin):
    """Range of lengths (mm)."""

    pass


class WithPressureMpaMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = PressureMPaStr.unit


class AmountPressureMpa(Amount, WithPressureMpaMixin):
    """Length (mm)."""

    pass


class AmountRangePressureMpa(AmountRangeWithDimensionality, WithPressureMpaMixin):
    """Range of lengths (mm)."""

    pass


class WithAreaM2Mixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = AreaM2Str.unit


class AmountAreaM2(Amount, WithAreaM2Mixin):
    """Area (m2)."""

    pass


class AmountRangeAreaM2(AmountRangeWithDimensionality, WithAreaM2Mixin):
    """Range of Area (m2)."""

    pass


class WithLengthInchStr(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = LengthInchStr.unit


class AmountLengthInch(Amount, WithLengthInchStr):
    """Length (inch)."""

    pass


class AmountRangeLengthInch(AmountRangeWithDimensionality, WithLengthInchStr):
    """Range of Length (inch)."""

    pass


class WithTemperatureCMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = TemperatureCStr.unit


class AmountTemperatureC(Amount, WithTemperatureCMixin):
    """Temperature (degrees C)."""

    pass


class AmountRangeTemperatureC(AmountRangeWithDimensionality, WithTemperatureCMixin):
    """Range of Temperature (degrees C)."""

    pass


class WithCapacityPerHourMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = CapacityPerHourStr.unit


class AmountCapacityPerHour(Amount, WithCapacityPerHourMixin):
    """Capacity per hour."""

    pass


class AmountRangeCapacityPerHour(AmountRangeWithDimensionality, WithCapacityPerHourMixin):
    """Capacity per hour range."""

    pass


class WithRValueMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = RValueStr.unit


class AmountRValue(Amount, WithRValueMixin):
    """R-Value."""

    pass


class AmountRangeRValue(AmountRangeWithDimensionality, WithRValueMixin):
    """R-Value range."""

    pass


class WithSpeedMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = SpeedStr.unit


class AmountSpeed(Amount, WithSpeedMixin):
    """Speed."""

    pass


class AmountRangeSpeed(AmountRangeWithDimensionality, WithSpeedMixin):
    """Speed range."""

    pass


class WithColorTemperatureMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = ColorTemperatureStr.unit


class AmountColorTemperature(Amount, WithColorTemperatureMixin):
    """Color temperature."""

    pass


class AmountRangeColorTemperature(AmountRangeWithDimensionality, WithColorTemperatureMixin):
    """Color temperature range."""

    pass


class WithLuminosityMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = LuminosityStr.unit


class AmountLuminosity(Amount, WithLuminosityMixin):
    """Luminosity."""

    pass


class AmountRangeLuminosity(AmountRangeWithDimensionality, WithLuminosityMixin):
    """Luminosity range."""

    pass


class WithPowerMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = PowerStr.unit


class AmountPower(Amount, WithPowerMixin):
    """Power."""

    pass


class AmountRangePower(AmountRangeWithDimensionality, WithPowerMixin):
    """Power range."""

    pass


class WithElectricalCurrentMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = ElectricalCurrentStr.unit


class AmountElectricalCurrent(Amount, WithElectricalCurrentMixin):
    """Current."""

    pass


class AmountRangeElectricalCurrent(AmountRangeWithDimensionality, WithElectricalCurrentMixin):
    """Current range."""

    pass


class WithVolumeMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = VolumeStr.unit


class AmountVolume(Amount, WithVolumeMixin):
    """Volume."""

    pass


class AmountRangeVolume(AmountRangeWithDimensionality, WithVolumeMixin):
    """Volume range."""

    pass


class WithAirflowMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = AirflowStr.unit


class AmountAirflow(Amount, WithAirflowMixin):
    """Airflow."""

    pass


class AmountRangeAirflow(AmountRangeWithDimensionality, WithAirflowMixin):
    """Airflow range."""

    pass


class WithFlowRateMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = FlowRateStr.unit


class AmountFlowRate(Amount, WithFlowRateMixin):
    """Flow Rate."""

    pass


class AmountRangeFlowRate(AmountRangeWithDimensionality, WithFlowRateMixin):
    """Flow Rate range."""

    pass


class WithMassPerLengthMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = MassPerLengthStr.unit


class AmountMassPerLength(Amount, WithFlowRateMixin):
    """Mass per length."""

    pass


class AmountRangeMassPerLength(AmountRangeWithDimensionality, WithFlowRateMixin):
    """Mass per length range."""

    pass


class WithAreaPerVolumeMixin(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = AreaPerVolumeStr.unit


class AmountAreaPerVolume(Amount, WithFlowRateMixin):
    """Area per volume."""

    pass


class AmountRangeAreaPerVolume(AmountRangeWithDimensionality, WithFlowRateMixin):
    """Area per volume range."""

    pass


class WithThermalConductivity(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = ThermalConductivityStr.unit


class AmountThermalConductivityMixin(Amount, WithThermalConductivity):
    """Area per volume."""

    pass


class AmountRangeThermalConductivity(AmountRangeWithDimensionality, WithThermalConductivity):
    """Area per volume range."""

    pass


class WithForce(WithDimensionalityMixin):
    """
    Unit validation mixin.

    May the Force be with you.
    """

    dimensionality_unit = ForceNStr.unit


class AmountForce(Amount, WithForce):
    """Area per volume."""

    pass


class AmountRangeForce(AmountRangeWithDimensionality, WithForce):
    """Area per volume range."""

    pass


class WithYarnWeight(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = YarnWeightStr.unit


class AmountYarnWeight(Amount, WithYarnWeight):
    """Yarn weight."""

    pass


class AmountRangeYarnWeight(AmountRangeWithDimensionality, WithYarnWeight):
    """Yarn weight range."""

    pass


class WithThermalExpansion(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = ThermalExpansionStr.unit


class AmountThermalExpansion(Amount, WithThermalExpansion):
    """Yarn weight."""

    pass


class AmountRangeThermalExpansion(AmountRangeWithDimensionality, WithThermalExpansion):
    """Yarn weight range."""

    pass


class WithUtilization(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = UtilizationStr.unit


class AmountUtilization(Amount, WithUtilization):
    """Utilization."""

    pass


class AmountRangeUtilization(AmountRangeWithDimensionality, WithUtilization):
    """Utilization range."""

    pass


class WithUFactor(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = UFactorStr.unit


class AmountUFactor(Amount, WithUFactor):
    """U-Factor."""

    pass


class AmountRangeUFactor(AmountRangeWithDimensionality, WithUFactor):
    """U-Factor range."""

    pass


class WithRFactor(WithDimensionalityMixin):
    """Unit validation mixin."""

    dimensionality_unit = RFactorStr.unit


class AmountRFactor(Amount, WithRFactor):
    """R-Factor."""

    pass


class AmountRangeRFactor(AmountRangeWithDimensionality, WithRFactor):
    """R-Factor range."""

    pass


# known range amounts
SUPPORTED_RANGE_TYPES: tuple[type[AmountRangeWithDimensionality], ...] = (
    AmountRangeMass,
    AmountRangeGWP,
    AmountRangeLengthM,
    AmountRangeLengthMm,
    AmountRangePressureMpa,
    AmountRangeAreaM2,
    AmountRangeLengthInch,
    AmountRangeTemperatureC,
    AmountRangeCapacityPerHour,
    AmountRangeRValue,
    AmountRangeSpeed,
    AmountRangeColorTemperature,
    AmountRangeLuminosity,
    AmountRangePower,
    AmountRangeElectricalCurrent,
    AmountRangeVolume,
    AmountRangeAirflow,
    AmountRangeFlowRate,
    AmountRangeMassPerLength,
    AmountRangeAreaPerVolume,
    AmountRangeThermalConductivity,
    AmountRangeForce,
    AmountRangeYarnWeight,
    AmountRangeThermalExpansion,
    AmountRangeUtilization,
    AmountRangeUFactor,
    AmountRangeRFactor,
)

# known range amount mapping by unit
SUPPORTED_RANGES_BY_UNIT: dict[str, type[AmountRangeWithDimensionality]] = {
    str(t.dimensionality_unit): t for t in SUPPORTED_RANGE_TYPES
}



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.api.dto.base import BaseOpenEpdApiModel
from openepd.compat.pydantic import pyd


class MaterialFilterDefinition(BaseOpenEpdApiModel):
    """
    Material filter definition.

    This object includes material filter itself as well as its hash.
    """

    mf: str = pyd.Field(
        title="MaterialFilter",
        example='!EC3 search("AluminiumBillets") !pragma oMF("1.0/1")',
        description="MaterialFilter in string format",
    )
    mf_hash: str = pyd.Field(
        title="MaterialFilter hash",
        example="22bf5b78cee5b79e1c76e818873d521c3972688b",
        description="MaterialFilter hash. Can be used to compare filters for equality, put to cache etc.",
    )


class MaterialFilterMeta(BaseOpenEpdApiModel):
    """Meta holding supplementary information about OMF query execution."""

    excluded_fields: list[str] | None = pyd.Field(
        example=["building_jurisdiction", "jurisdiction"],
        description="list of fields excluded by server process for any reason",
        default=None,
    )
    effective_omf: MaterialFilterDefinition = pyd.Field(
        description="Effective OpenMaterialFilter as applied to search, after transformations if any"
    )


class MaterialFilterMetaMixin(BaseOpenEpdApiModel):
    """Mixing for adding MaterialFilterMeta to MetaCollection."""

    material_filter: MaterialFilterMeta | None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import CableTraysMaterial, EnergySource, RacewaysMaterial
from openepd.model.specs.singular.mixins.conduit_mixin import ConduitMixin
from openepd.model.validation.quantity import (
    ColorTemperatureStr,
    LengthMmStr,
    LengthMStr,
    LuminosityStr,
    MassKgStr,
    PowerStr,
    UtilizationStr,
    validate_quantity_ge_factory,
    validate_quantity_le_factory,
    validate_quantity_unit_factory,
)


class LowVoltBusesV1(BaseOpenEpdHierarchicalSpec):
    """Busbars and Busways of 600V or less."""

    _EXT_VERSION = "1.0"


class MedVoltBusesV1(BaseOpenEpdHierarchicalSpec):
    """Busbars and Busways over 600V."""

    _EXT_VERSION = "1.0"


class BatteriesV1(BaseOpenEpdHierarchicalSpec):
    """Battery equipment, including central batteries, battery charging, and UPS."""

    _EXT_VERSION = "1.0"


class OtherElectricalPowerStorageV1(BaseOpenEpdHierarchicalSpec):
    """Other electrical power storage performance specification."""

    _EXT_VERSION = "1.0"


class CableTraysV1(BaseOpenEpdHierarchicalSpec):
    """Mechanical support for electrial or communications cabling, typically suspended from a roof or wall."""

    _EXT_VERSION = "1.0"

    # Own fields:
    height: LengthMmStr | None = pyd.Field(default=None, description="", example="100 mm")
    width: LengthMmStr | None = pyd.Field(default=None, description="", example="100 mm")
    depth: LengthMmStr | None = pyd.Field(default=None, description="", example="100 mm")
    static_load: MassKgStr | None = pyd.Field(default=None, description="", example="1 kg")
    ventilated: bool | None = pyd.Field(
        default=None, description="At least 40% of the tray base is open to air flow", example=True
    )
    cable_trays_material: CableTraysMaterial | None = pyd.Field(default=None, description="", example="Stainless Steel")


class ElectricalBusesV1(BaseOpenEpdHierarchicalSpec):
    """
    Power distribution, in the form of busbars or of insulted ducts made of copper or aluminum busbars.

    It is an alternative means of conducting electricity compared toto power cables or cable bus. Also called
    bus ducts.
    """

    _EXT_VERSION = "1.0"

    # Nested specs:
    LowVoltBuses: LowVoltBusesV1 | None = None
    MedVoltBuses: MedVoltBusesV1 | None = None


class FloorEquipmentBoxesV1(BaseOpenEpdHierarchicalSpec):
    """Equipment boxes for power or electronic equipment embedded in an accessible floor."""

    _EXT_VERSION = "1.0"


class PowerDistributionUnitsV1(BaseOpenEpdHierarchicalSpec):
    """Switched electrical distribution units placed very close to the point of consumption, for example inside a rack of electronic equipment."""

    _EXT_VERSION = "1.0"


class RacewaysV1(BaseOpenEpdHierarchicalSpec):
    """Mechanical guideways for eletrical communications cabling, typically embedded in an accessible floor."""

    _EXT_VERSION = "1.0"

    # Own fields:
    width: LengthMStr | None = pyd.Field(default=None, description="", example="100 mm")
    depth: LengthMStr | None = pyd.Field(default=None, description="", example="100 mm")
    painted: bool | None = pyd.Field(default=None, description="", example=True)
    divided: bool | None = pyd.Field(default=None, description="", example=True)
    raceways_material: RacewaysMaterial | None = pyd.Field(default=None, description="", example="Aluminum")


class FueledElectricalGeneratorsV1(BaseOpenEpdHierarchicalSpec):
    """Fueled electrical generators."""

    _EXT_VERSION = "1.0"


class OtherGenerationV1(BaseOpenEpdHierarchicalSpec):
    """Other generation."""

    _EXT_VERSION = "1.0"


class PhotovoltaicsV1(BaseOpenEpdHierarchicalSpec):
    """Solar photovoltaics, rated on a nameplate capacity basis."""

    _EXT_VERSION = "1.0"


class WindTurbinesV1(BaseOpenEpdHierarchicalSpec):
    """Wind generators, rated on a nameplate capacity basis."""

    _EXT_VERSION = "1.0"


class ElectricityFromPowerGridV1(BaseOpenEpdHierarchicalSpec):
    """Electrical energy drawn from a specific utility grid."""

    _EXT_VERSION = "1.0"


class ElectricityFromSpecificGeneratorV1(BaseOpenEpdHierarchicalSpec):
    """Electrical energy from a specific power plant, such as a wind farm using a specific type of turbine."""

    _EXT_VERSION = "1.0"

    # Own fields:
    energy_source: EnergySource | None = pyd.Field(default=None, description="", example="Grid")


class PowerPurchaseAgreementsV1(BaseOpenEpdHierarchicalSpec):
    """
    Electrical energy subject to a verified power purchase agreement.

    The impact of electricity generation is allocated specifically to the agreement and not to the general grid.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    energy_source: EnergySource | None = pyd.Field(default=None, description="", example="Grid")


class LightbulbsV1(BaseOpenEpdHierarchicalSpec):
    """Various types of light bulbs, including LED, CFL, halogen, and incandescent."""

    _EXT_VERSION = "1.0"


class LightingControlsV1(BaseOpenEpdHierarchicalSpec):
    """Devices used to control the operation of lighting, including dimmers, sensors, and smart controls."""

    _EXT_VERSION = "1.0"


class LightingFixturesV1(BaseOpenEpdHierarchicalSpec):
    """Permanent lighting fixtures for interior spaces, including ceiling, wall-mounted, and pendant fixtures."""

    _EXT_VERSION = "1.0"


class OutdoorLightingV1(BaseOpenEpdHierarchicalSpec):
    """Lighting products designed for outdoor use, including landscape and security lighting."""

    _EXT_VERSION = "1.0"


class SpecialtyLightingV1(BaseOpenEpdHierarchicalSpec):
    """Specialized lighting for niche applications like emergency, medical, or theatrical lighting."""

    _EXT_VERSION = "1.0"


class TaskLightingV1(BaseOpenEpdHierarchicalSpec):
    """Lighting designed for specific tasks such as desk lamps, under-cabinet lighting, and reading lamps."""

    _EXT_VERSION = "1.0"


class ElectricalPowerStorageV1(BaseOpenEpdHierarchicalSpec):
    """Electrical Power Storage."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    Batteries: BatteriesV1 | None = None
    OtherElectricalPowerStorage: OtherElectricalPowerStorageV1 | None = None


class LowVoltageElectricalDistributionV1(BaseOpenEpdHierarchicalSpec):
    """Low Voltage Electrical Distribution."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    CableTrays: CableTraysV1 | None = None
    ElectricalBuses: ElectricalBusesV1 | None = None
    FloorEquipmentBoxes: FloorEquipmentBoxesV1 | None = None
    PowerDistributionUnits: PowerDistributionUnitsV1 | None = None
    Raceways: RacewaysV1 | None = None


class ElectricalGenerationEquipmentV1(BaseOpenEpdHierarchicalSpec):
    """
    Equipment for generating electrical power.

    This category is primarily for smaller-scale. (e.g. on premises) generation, rather than utility-scale equipment.
    """

    _EXT_VERSION = "1.0"

    # Nested specs:
    FueledElectricalGenerators: FueledElectricalGeneratorsV1 | None = None
    OtherGeneration: OtherGenerationV1 | None = None
    Photovoltaics: PhotovoltaicsV1 | None = None
    WindTurbines: WindTurbinesV1 | None = None


class ElectricPowerV1(BaseOpenEpdHierarchicalSpec):
    """Electrical energy drawn from a utility grid."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    ElectricityFromPowerGrid: ElectricityFromPowerGridV1 | None = None
    ElectricityFromSpecificGenerator: ElectricityFromSpecificGeneratorV1 | None = None
    PowerPurchaseAgreements: PowerPurchaseAgreementsV1 | None = None


class LightingV1(BaseOpenEpdHierarchicalSpec):
    """Lamps and lightbulbs and lamp components."""

    _EXT_VERSION = "1.0"

    # Own fields:
    color_temperature: ColorTemperatureStr | None = pyd.Field(default=None, description="", example="1 K")
    typical_utilization: UtilizationStr | None = pyd.Field(default=None, description="", example="1 h / yr")
    luminosity: LuminosityStr | None = pyd.Field(default=None, description="", example="1 lumen")
    wattage: PowerStr | None = pyd.Field(default=None, description="")
    color_rendering_index: float | None = pyd.Field(default=None, description="", example=2.3)
    dimmable: bool | None = pyd.Field(default=None, description="", example=True)

    _color_temperature_quantity_ge_validator = pyd.validator("color_temperature", allow_reuse=True)(
        validate_quantity_ge_factory("1E+03 K")
    )
    _color_temperature_quantity_le_validator = pyd.validator("color_temperature", allow_reuse=True)(
        validate_quantity_le_factory("1E+04 K")
    )
    _typical_utilization_unit_validator = pyd.validator("typical_utilization", allow_reuse=True)(
        validate_quantity_unit_factory("h / yr")
    )
    _typical_utilization_quantity_ge_validator = pyd.validator("typical_utilization", allow_reuse=True)(
        validate_quantity_ge_factory("25 h / yr")
    )
    _luminosity_quantity_ge_validator = pyd.validator("luminosity", allow_reuse=True)(
        validate_quantity_ge_factory("450 lumen")
    )
    _luminosity_quantity_le_validator = pyd.validator("luminosity", allow_reuse=True)(
        validate_quantity_le_factory("2.6E+03 lumen")
    )
    _wattage_quantity_ge_validator = pyd.validator("wattage", allow_reuse=True)(validate_quantity_ge_factory("5 W"))
    _wattage_quantity_le_validator = pyd.validator("wattage", allow_reuse=True)(validate_quantity_le_factory("100 W"))

    # Nested specs:
    Lightbulbs: LightbulbsV1 | None = None
    LightingControls: LightingControlsV1 | None = None
    LightingFixtures: LightingFixturesV1 | None = None
    OutdoorLighting: OutdoorLightingV1 | None = None
    SpecialtyLighting: SpecialtyLightingV1 | None = None
    TaskLighting: TaskLightingV1 | None = None


class ElectricalConduitV1(BaseOpenEpdHierarchicalSpec, ConduitMixin):
    """Tubing used to protect and route electrical wiring in a building or structure."""

    _EXT_VERSION = "1.1"


class ElectricalV1(BaseOpenEpdHierarchicalSpec):
    """Electric power and equipment."""

    _EXT_VERSION = "1.1"

    # Nested specs:
    ElectricalPowerStorage: ElectricalPowerStorageV1 | None = None
    LowVoltageElectricalDistribution: LowVoltageElectricalDistributionV1 | None = None
    ElectricalGenerationEquipment: ElectricalGenerationEquipmentV1 | None = None
    ElectricPower: ElectricPowerV1 | None = None
    Lighting: LightingV1 | None = None
    ElectricalConduit: ElectricalConduitV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import (
    AccessFlooringCoreMaterial,
    AccessFlooringFinishMaterial,
    AccessFlooringSeismicRating,
    AccessFlooringStringers,
    AllFabrication,
    CarpetFormFactor,
    CarpetIntendedApplication,
    CarpetManufactureType,
    CarpetYarnType,
    CeilingPanelCoreMaterial,
    CeilingPanelFireRating,
    CementBoardThickness,
    DeckingBoardMaterial,
    GypsumFacing,
    GypsumFireRating,
    GypsumThickness,
    PlasterComposition,
    ResilientFlooringFormFactor,
    ResilientFlooringMaterial,
    ResilientFlooringThickness,
    SawnTimberSpecies,
    VinylSheetConstruction,
    WallBaseMaterial,
    WoodFlooringFabrication,
    WoodFlooringTimberSpecies,
)
from openepd.model.specs.singular.common import HasForestPracticesCertifiers
from openepd.model.validation.numbers import RatioFloat
from openepd.model.validation.quantity import (
    AreaPerVolumeStr,
    ForceNStr,
    GwpKgCo2eStr,
    LengthMmStr,
    LengthMStr,
    PressureMPaStr,
    RFactorStr,
    YarnWeightStr,
    validate_quantity_ge_factory,
    validate_quantity_le_factory,
    validate_quantity_unit_factory,
)


class AccessFlooringV1(BaseOpenEpdHierarchicalSpec):
    """
    Elevated floor system built on top of concrete slab surface.

    It thereby creates a hidden void between the two floors that is used for the passage of mechanical and electrical
    services. The system consists of panels, stringers, and pedestals.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    core_material: AccessFlooringCoreMaterial | None = pyd.Field(default=None, description="", example="Cementitious")
    finish_material: AccessFlooringFinishMaterial | None = pyd.Field(default=None, description="", example="Linoleum")
    stringers: AccessFlooringStringers | None = pyd.Field(default=None, description="", example="Standard")
    seismic_rating: AccessFlooringSeismicRating | None = pyd.Field(default=None, description="", example="Type 0")
    magnetically_attached_finish: bool | None = pyd.Field(default=None, description="", example=True)
    permanent_finish: bool | None = pyd.Field(default=None, description="", example=True)
    drylay: bool | None = pyd.Field(default=None, description="", example=True)
    adjustable_height: bool | None = pyd.Field(default=None, description="", example=True)
    fixed_height: bool | None = pyd.Field(default=None, description="", example=True)
    finished_floor_height: LengthMmStr | None = pyd.Field(default=None, description="", example="1 m")
    panel_thickness: LengthMmStr | None = pyd.Field(default=None, description="", example="1 m")
    concentrated_load: PressureMPaStr | None = pyd.Field(default=None, description="", example="1 MPa")
    uniform_load: PressureMPaStr | None = pyd.Field(default=None, description="", example="1 MPa")
    rolling_load_10_pass: ForceNStr | None = pyd.Field(default=None, description="", example="1 N")
    rolling_load_10000_pass: ForceNStr | None = pyd.Field(default=None, description="", example="1 N")

    _access_flooring_rolling_load_10_pass_is_quantity_validator = pyd.validator(
        "rolling_load_10_pass", allow_reuse=True
    )(validate_quantity_unit_factory("N"))
    _access_flooring_rolling_load_10000_pass_is_quantity_validator = pyd.validator(
        "rolling_load_10000_pass", allow_reuse=True
    )(validate_quantity_unit_factory("N"))


class CarpetV1(BaseOpenEpdHierarchicalSpec):
    """Textile Floor Coverings."""

    _EXT_VERSION = "1.0"

    # Own fields:
    length: LengthMStr | None = pyd.Field(default=None, description="", example="1 m")
    width: LengthMStr | None = pyd.Field(default=None, description="", example="1 m")
    intended_application: list[CarpetIntendedApplication] | None = pyd.Field(
        default=None, description="", example=["Res"]
    )
    manufacture_type: CarpetManufactureType | None = pyd.Field(default=None, description="", example="Tufted")
    form_factor: CarpetFormFactor | None = pyd.Field(default=None, description="", example="Tiles")
    yarn_weight: YarnWeightStr | None = pyd.Field(default=None, description="", example="1 g / m2")
    yarn_type: CarpetYarnType | None = pyd.Field(default=None, description="", example="Nylon 6,6")
    fire_radiant_panel_rating_astme648: str | None = pyd.Field(default=None, description="")
    fire_smoke_density_rating_astme648: str | None = pyd.Field(default=None, description="")
    voc_emissions: str | None = pyd.Field(default=None, description="", example="test_valueValidatedStringProperty")
    cushioned: bool | None = pyd.Field(default=None, description="", example=True)
    bleachable: bool | None = pyd.Field(default=None, description="", example=True)
    gwp_factor_base: GwpKgCo2eStr | None = pyd.Field(default=None, description="", example="1 kgCO2e")
    gwp_factor_yarn: GwpKgCo2eStr | None = pyd.Field(default=None, description="", example="1 kgCO2e")

    _yarn_weight_is_quantity_validator = pyd.validator("yarn_weight", allow_reuse=True)(
        validate_quantity_unit_factory("g / m2")
    )
    _yarn_weight_ge_validator = pyd.validator("yarn_weight", allow_reuse=True)(validate_quantity_ge_factory("0 g / m2"))


class LaminateV1(BaseOpenEpdHierarchicalSpec):
    """Laminate flooring."""

    _EXT_VERSION = "1.0"


class OtherFlooringV1(BaseOpenEpdHierarchicalSpec):
    """Other not yet classified kinds of flooring."""

    _EXT_VERSION = "1.0"


class ResilientFlooringV1(BaseOpenEpdHierarchicalSpec):
    """
    Resilient floor products.

    Includes vinyl, rubber, linoleum, composition cork, etc. in modular square or rectangle shapes.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    length: LengthMStr | None = pyd.Field(default=None, description="", example="1 m")
    width: LengthMStr | None = pyd.Field(default=None, description="", example="1 m")
    form_factor: ResilientFlooringFormFactor | None = pyd.Field(default=None, description="", example="Loose Lay")
    material: ResilientFlooringMaterial | None = pyd.Field(default=None, description="", example="VCT")
    sheet_construction: VinylSheetConstruction | None = pyd.Field(default=None, description="", example="Homogeneous")
    wear_layer: LengthMmStr | None = pyd.Field(default=None, description="", example="1 m")
    delta_iic: float | None = pyd.Field(default=None, description="", example=2.3)
    thickness: ResilientFlooringThickness | None = pyd.Field(default=None, description="", example="≤ 2mm")
    sport_flooring: bool | None = pyd.Field(default=None, description="", example=True)
    conductive_flooring: bool | None = pyd.Field(default=None, description="", example=True)
    zwtl: bool | None = pyd.Field(default=None, description="", example=True)
    floor_score: bool | None = pyd.Field(default=None, description="", example=True)
    nsf332: bool | None = pyd.Field(default=None, description="", example=True)


class WallBaseV1(BaseOpenEpdHierarchicalSpec):
    """Wall base made to help cover gaps between wall and vinyl, rubber, wood, or tile flooring."""

    _EXT_VERSION = "1.0"

    # Own fields:
    wall_base_material: WallBaseMaterial | None = pyd.Field(default=None, description="", example="Rubber")


class WoodFlooringV1(BaseOpenEpdHierarchicalSpec, HasForestPracticesCertifiers):
    """
    Wood flooring for interior applications.

    Includes hardwood strip and plank flooring, engineered hardwood flooring, wood parquet flooring, coordinated
    transitions, and molding pieces.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    thickness: LengthMmStr | None = pyd.Field(default=None, description="", example="10 mm")
    timber_species: WoodFlooringTimberSpecies | None = pyd.Field(default=None, description="", example="Oak")
    fabrication: WoodFlooringFabrication | None = pyd.Field(default=None, description="", example="Solid hardwood")


class AcousticalCeilingsV1(BaseOpenEpdHierarchicalSpec):
    """Acoustical ceiling panels."""

    _EXT_VERSION = "1.0"

    # Own fields:
    thickness: LengthMmStr | None = pyd.Field(default=None, description="", example="10 mm")


class CeramicTileV1(BaseOpenEpdHierarchicalSpec):
    """Ceramic tiles, including porcelain, quarry, pressed floor tile, wall tile, mosaic tile, etc."""

    _EXT_VERSION = "1.0"

    # Own fields:
    porcelain: bool | None = pyd.Field(
        default=None, description="A dense and durable ceramic tile made from fine porcelain clay.", example=True
    )
    quarry: bool | None = pyd.Field(
        default=None,
        description="A type of unglazed ceramic tile made from natural clay with a slightly rough texture.",
        example=True,
    )
    pressed_floor_tile: bool | None = pyd.Field(
        default=None,
        description="A durable and low-maintenance type of tile made by compressing clay or "
        "other materials at high pressure.",
        example=True,
    )
    wall_tile: bool | None = pyd.Field(
        default=None,
        description="A decorative tile designed for use on vertical surfaces such as walls or backsplashes.",
        example=True,
    )
    mosaic_tile: bool | None = pyd.Field(
        default=None,
        description="A small decorative tile made of glass, stone, or ceramic, arranged in a pattern "
        "to create a design.",
        example=True,
    )
    specialty: bool | None = pyd.Field(
        default=None,
        description="A unique and customized type of tile, often made from unconventional materials or with "
        "specialized designs or finishes.",
        example=True,
    )


class GaugedTileV1(BaseOpenEpdHierarchicalSpec):
    """
    Specially manufactured porcelain tile with extra-large panels/slabs.

    Manufactured to a specific thickness ranging from 2-20 mm.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    tile_panels: bool | None = pyd.Field(
        default=None,
        description="Large-format porcelain or natural stone tiles that are typically over 3 feet in length and width, "
        "designed for use in floor and wall installations to create a seamless and uninterrupted "
        "appearance.",
        example=True,
    )
    tile_pavers: bool | None = pyd.Field(
        default=None,
        description="Thick and durable porcelain or natural stone tiles that are commonly used in outdoor "
        "applications, such as patios, walkways, and driveways, due to their high resistance to weather "
        "and wear.",
        example=True,
    )


class GlassTileV1(BaseOpenEpdHierarchicalSpec):
    """Glass Tiles."""

    _EXT_VERSION = "1.0"

    # Own fields:
    regular: bool | None = pyd.Field(
        default=None,
        description="Glass tile that is typically square or rectangular in shape, and used for a variety of "
        "decorative applications, such as kitchen backsplashes, shower walls, and accent borders.",
        example=True,
    )
    glass_mosaic: bool | None = pyd.Field(
        default=None,
        description="A small, decorative glass tile made in a variety of shapes and colors, used for intricate "
        "designs and patterns on walls, floors, and other surfaces.",
        example=True,
    )
    miniature_mosaic: bool | None = pyd.Field(
        default=None,
        description="Glass mosaic tile that is smaller in size than regular glass mosaic tile, often used for "
        "intricate details and designs in backsplashes, shower walls, and decorative accents.",
        example=True,
    )
    large_format: bool | None = pyd.Field(
        default=None,
        description="Glass tile that is larger in size than regular glass tile, often used to create a dramatic and "
        "modern effect in commercial and residential spaces.",
        example=True,
    )


class GypsumSupportsV1(BaseOpenEpdHierarchicalSpec):
    """Supports for suspended and furred gypsum wall and ceiling assemblies."""

    _EXT_VERSION = "1.0"


class FlooringV1(BaseOpenEpdHierarchicalSpec):
    """General category - finishes for floors."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    AccessFlooring: AccessFlooringV1 | None = None
    Carpet: CarpetV1 | None = None
    Laminate: LaminateV1 | None = None
    OtherFlooring: OtherFlooringV1 | None = None
    ResilientFlooring: ResilientFlooringV1 | None = None
    WallBase: WallBaseV1 | None = None
    WoodFlooring: WoodFlooringV1 | None = None


class CeilingPanelV1(BaseOpenEpdHierarchicalSpec):
    """Acoustical and other specialty ceiling panels."""

    _EXT_VERSION = "1.0"
    """Modular Ceiling Panels"""

    # Own fields:
    fire_rating: CeilingPanelFireRating | None = pyd.Field(default=None, description="", example="Class A")
    core_material: CeilingPanelCoreMaterial | None = pyd.Field(default=None, description="", example="Fiberglass")
    nrc: RatioFloat | None = pyd.Field(
        default=None,
        description="Noise Reduction Coefficient (NRC) or Sound Absorbtion Average (SAA) per ASTM C423",
        example=0.5,
        ge=0,
        le=1,
    )
    cac: int | None = pyd.Field(
        default=None, description="Ceiling Attenuation Class (CAC) per ASTM E1414", example=13, ge=10, le=50
    )

    # Nested specs:
    AcousticalCeilings: AcousticalCeilingsV1 | None = None


class BackingAndUnderlayV1(BaseOpenEpdHierarchicalSpec):
    """Cementitious, glass-mat faced gypsum, and fibered gypsum backing boards to support finish materials."""

    _EXT_VERSION = "1.0"


class CementBoardV1(BaseOpenEpdHierarchicalSpec):
    """Hard cementitious boards, typically used as a tile backer."""

    _EXT_VERSION = "1.0"

    # Own fields:
    thickness: CementBoardThickness | None = pyd.Field(
        default=None, description="", example=str(CementBoardThickness.INCH_1_2)
    )

    _cement_board_thickness_is_quantity_validator = pyd.validator("thickness", allow_reuse=True)(
        validate_quantity_unit_factory("m")
    )


class TilingV1(BaseOpenEpdHierarchicalSpec):
    """
    Decorative building materials that includes a variety of ceramic, porcelain, and glass tiles.

    Used for covering and enhancing surfaces such as floors, walls, and countertops.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    thickness: LengthMmStr | None = pyd.Field(default=None, description="", example="9 mm")
    flooring: bool | None = pyd.Field(default=None, description="Tiling intended for walking.", example=True)
    wall_finish: bool | None = pyd.Field(
        default=None,
        description="A decorative tile designed for use on vertical surfaces such as walls or backsplashes.",
        example=True,
    )
    cladding: bool | None = pyd.Field(
        default=None,
        description="Tiling for exterior use, primarily used for the walls of buildings and structures, providing a "
        "protective and decorative layer that enhances the aesthetic appearance and weather resistance "
        "of the underlying structure.",
        example=True,
    )
    other: bool | None = pyd.Field(
        default=None, description="Tiling used as countertops, ceilings, furnishings, hardscapes etc.", example=True
    )
    residential_only: bool | None = pyd.Field(
        default=None,
        description="All commercial tile can also be used in residential applications, but the opposite may not be "
        "true. This selection allows to filter out tiling that is not intended for commercial "
        "applications.",
        example=True,
    )
    reinforced: bool | None = pyd.Field(
        default=None,
        description="Steel-reinforced ceramic tiles or tiles with other special reinforcing technology.",
        example=True,
    )
    total_recycled_content: RatioFloat | None = pyd.Field(
        default=None,
        description="Proportion of this product that is sourced from recycled content. Pre-consumer recycling is "
        "given a 50% weighting, 100% for post-consumer, by mass.",
        example=0.5,
        ge=0,
        le=1,
    )
    post_consumer_recycled_content: RatioFloat | None = pyd.Field(
        default=None,
        description="Proportion of this product that is sourced from post-consumer recycled content, by mass.",
        example=0.5,
        ge=0,
        le=1,
    )

    _thickness_max_validator = pyd.validator("thickness", allow_reuse=True)(validate_quantity_le_factory("50 mm"))

    # Nested specs:
    CeramicTile: CeramicTileV1 | None = None
    GaugedTile: GaugedTileV1 | None = None
    GlassTile: GlassTileV1 | None = None


class DeckingBoardsV1(BaseOpenEpdHierarchicalSpec, HasForestPracticesCertifiers):
    """Decking boards provide the finished surface of a deck and support the weight of people and furniture."""

    _EXT_VERSION = "1.0"

    # Own fields:
    timber_species: SawnTimberSpecies | None = pyd.Field(default=None, description="", example="Alaska Cedar")
    fabrication: AllFabrication | None = pyd.Field(default=None, description="", example="LVL")
    weather_exposed: bool | None = pyd.Field(default=None, description="", example=True)
    fire_retardant: bool | None = pyd.Field(default=None, description="", example=True)
    decay_resistant: bool | None = pyd.Field(default=None, description="", example=True)
    material: DeckingBoardMaterial | None = pyd.Field(default=None, description="", example="Wood")


class GlassFiberReinforcedGypsumFabricationsV1(BaseOpenEpdHierarchicalSpec):
    """Gypsum with integrated glass fiber reinforcement, which may be fabricated in complex shapes or as a board."""

    _EXT_VERSION = "1.0"


class GypsumV1(BaseOpenEpdHierarchicalSpec):
    """Gypsum board used for interior walls, ceilings, and similar applications."""

    _EXT_VERSION = "1.1"

    # Own fields:
    fire_rating: GypsumFireRating | None = pyd.Field(default=None, description="", example="-")
    thickness: GypsumThickness | None = pyd.Field(default=None, description="", example="1 m")
    facing: GypsumFacing | None = pyd.Field(default=None, description="", example="Paper")
    r_factor: RFactorStr | None = pyd.Field(default=None, description="", example="1 RSI")
    flame_spread_astm_e84: int | None = pyd.Field(default=None, description="", example=3)
    smoke_production_astm_e84: int | None = pyd.Field(default=None, description="", example=3)
    surface_abrasion_d4977: int | None = pyd.Field(default=None, description="", example=3)
    indentation_d5420: int | None = pyd.Field(default=None, description="", example=3)
    soft_body_impact_e695: int | None = pyd.Field(default=None, description="", example=3)
    hard_body_impact_c1929: int | None = pyd.Field(default=None, description="", example=3)
    mold_resistant: bool | None = pyd.Field(default=None, description="", example=True)
    foil_backing: bool | None = pyd.Field(default=None, description="", example=True)
    moisture_resistant: bool | None = pyd.Field(default=None, description="", example=True)
    abuse_resistant: bool | None = pyd.Field(default=None, description="", example=True)

    _gypsum_r_factor_is_quantity_validator = pyd.validator("r_factor", allow_reuse=True)(
        validate_quantity_unit_factory("RSI")
    )

    # Nested specs:
    GypsumSupports: GypsumSupportsV1 | None = None


class MirrorsV1(BaseOpenEpdHierarchicalSpec):
    """Mirrors."""

    _EXT_VERSION = "1.0"


class PaintByMassV1(BaseOpenEpdHierarchicalSpec):
    """
    Paintings and coatings by mass.

    Expected declared unit for products is mass of ready to use product.
    """

    _EXT_VERSION = "1.0"


class PaintByVolumeV1(BaseOpenEpdHierarchicalSpec):
    """
    Paintings and coatings by volume.

    Expecting declared unit for products is volume of ready to use product.
    """

    _EXT_VERSION = "1.0"


class PaintByAreaV1(BaseOpenEpdHierarchicalSpec):
    """
    Paintings and coatings by area.

    Expected declared unit is area of host surface covered by the product.
    """

    _EXT_VERSION = "1.0"


class PaintingAndCoatingV1(BaseOpenEpdHierarchicalSpec):
    """Paintings and coatings."""

    _EXT_VERSION = "1.1"

    PaintByMass: PaintByMassV1 | None = None
    PaintByVolume: PaintByVolumeV1 | None = None
    PaintByArea: PaintByAreaV1 | None = None


class WallFinishesV1(BaseOpenEpdHierarchicalSpec):
    """Interior wall coverings including fabric, textile, wood, stone, and metal products."""

    _EXT_VERSION = "1.0"

    # Own fields:
    thickness: LengthMmStr | None = pyd.Field(default=None, description="", example="10 mm")


class PlasterV1(BaseOpenEpdHierarchicalSpec):
    """
    Plaster, Stucco, & Render.

    Used for the protective or decorative coating of walls and ceilings and for
    moulding and casting decorative elements. These are typically gypsum-, lime-,
    or cement-based. Products in this category refer to dry mix.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    composition: PlasterComposition | None = pyd.Field(default=None, description="", example="Cement")
    application_rate: AreaPerVolumeStr | None = pyd.Field(
        default=None,
        description="Typical or reference amount of material covering a unit of a host surface.",
        example="10 m2/l",
    )


class FinishesV1(BaseOpenEpdHierarchicalSpec):
    """General category - finishes for interior ceilings, floors, walls."""

    _EXT_VERSION = "1.1"

    # Nested specs:
    Flooring: FlooringV1 | None = None
    CeilingPanel: CeilingPanelV1 | None = None
    BackingAndUnderlay: BackingAndUnderlayV1 | None = None
    CementBoard: CementBoardV1 | None = None
    Tiling: TilingV1 | None = None
    DeckingBoards: DeckingBoardsV1 | None = None
    GlassFiberReinforcedGypsumFabrications: GlassFiberReinforcedGypsumFabricationsV1 | None = None
    Gypsum: GypsumV1 | None = None
    Mirrors: MirrorsV1 | None = None
    PaintingAndCoating: PaintingAndCoatingV1 | None = None
    WallFinishes: WallFinishesV1 | None = None
    Plaster: PlasterV1 | None = None
